from typing import Optional, Union, Dict
from collections.abc import Iterable
from abc import ABC, abstractmethod

import tensorflow as tf
import numpy as np

class AbstractModel(ABC):
    """Abstract class for building quantum models"""
    def __init__(self, name:Optional[str]=None,
                 random_state:Optional[int]=None, 
                 checkpoint_dir:Optional[str]=None,
                 checkpoint_interval:int=10,
                 checkpoint_max_to_keep:Optional[int]=None):
        """Instantiate the quantum model
        
            Arguments:
                name: (Optional) string
                    Name given to the model.
                random_state: (Optional) int
                    The random state used at the beginning of a training
                    for reproducible result.
                checkpoint_dir: Optional[string]
                    The path to a directory in which to write checkpoints. If None,
                    no checkpoint will be saved.
                checkpoint_interval: int, default=10
                    Number of epochs between each checkpoint.
                checkpoint_max_to_keep: (Optional) int
                    Number of checkpoints to keep. If None, all checkpoints are kept.
        """
        # for restoring epoch at checkpoint
        self.start_epoch = 0
        
        self.name = name
        self.random_state = random_state
        
        self.set_random_state(self.random_state)
        self._setup_checkpoint(checkpoint_dir, checkpoint_interval,
                               checkpoint_max_to_keep)
        
        self._validate_init()
        
    def _validate_init(self):
        pass
    
    @abstractmethod
    def _create_checkpoint(self, *args, **kwargs):
        pass
    
    def _setup_checkpoint(self, directory:Optional[str]=None, interval:int=10,
                          max_to_keep:Optional[int]=None, *args, **kwargs):
        """Setup training checkpoints.
            
            Note that checkpoints are created only when training starts.
            
            Arguments:
                directory: string
                    The path to a directory in which to write checkpoints.
                interval: int, default=10
                    Number of epochs between each checkpoint.
                max_to_keep: (Optional) int
                    Number of checkpoints to keep. If None, all checkpoints are kept.
        """
        self.checkpoint_dir = directory
        self.checkpoint_interval = interval
        self.checkpoint_max_to_keep = max_to_keep
        
        if self.checkpoint_dir is not None:
            self.checkpoint = self._create_checkpoint(*args, **kwargs)
            self.checkpoint_manager = tf.train.CheckpointManager(self.checkpoint, 
                                                                 self.checkpoint_dir,
                                                                 self.checkpoint_max_to_keep)
        else:
            self.checkpoint = None
            self.checkpoint_manager = None
    
    @staticmethod
    def set_random_state(random_state:Optional[int]=None):
        """Set the global random states for the tensorflow and the numpy libraries
            
            Arguments:
                random_state: (Optional) int
                    Random state to set.
        """
        tf.random.set_seed(random_state)
        np.random.seed(random_state)
    
    @staticmethod
    def prepare_dataset(*data, batch_size:int=32, seed=None, drop_remainder=True, buffer_size:Optional[int]=None):
        """Create batched dataset with shuffling
        
            Arguments:
                *data: iterable of list/np.ndarray/tf.Tensor
                    Iterable of data that make up the dataset.
                batchsize: int
                    Number of consecutive elements of the dataset to combine in a single batch.
                seed: (Optional) int
                    Random seed used in shuffling the dataset.
                drop_remainder: boolean, default=True
                    Whether the last batch should be dropped in the case it has fewer than batch_size elements.
                buffer_size: (Optional) int
                    Buffer size used in shuffling the dataset. If None, the sample size is used as the buffer size.
            Returns:
                tf.Dataset containing the shuffled dataset.
        """        
        data_list = []
        data_sizes = []
        for d in data:
            data_list.append(tf.data.Dataset.from_tensor_slices(d))
            if isinstance(d, tuple):
                data_sizes += [d_.shape[0] for d_ in d]
            else:
                data_sizes.append(d.shape[0])
        if len(set(data_sizes)) > 1:
            raise ValueError("input data have inconsistent sample sizes")
        if buffer_size is None:
            buffer_size = data_sizes[0]
        ds = tf.data.Dataset.zip(tuple(data_list))
        ds = ds.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=True)
        ds = ds.batch(batch_size, drop_remainder=drop_remainder)
        ds = ds.prefetch(tf.data.AUTOTUNE)
        return ds
    
    @tf.function
    def to_prob(self, x):
        """Convert quantum circuit output to probabilities"""
        return tf.divide(tf.add(x, 1), 2)
    
    def restore_checkpoint(self, checkpoint:tf.train.Checkpoint=None):
        if (self.checkpoint is None) or (self.checkpoint_manager is None):
            raise RuntimeError("check point not initialized")
        if checkpoint is None:
            checkpoint = self.checkpoint_manager.latest_checkpoint
        self.checkpoint.restore(checkpoint)
        self.start_epoch = int(self.checkpoint.step) + 1
    
    def _train_preprocess(self, *args, **kwargs):
        self.set_random_state(self.random_state)
        
    def _train_post_epoch(self, epoch:int, *args, **kwargs):
        if (self.checkpoint is not None) and (self.checkpoint_manager is not None):
            self.checkpoint.step.assign_add(1)
            if int(self.checkpoint.step) % self.checkpoint_interval == 0:
                self.checkpoint_manager.save()
        
    def _train_postprocess(self, *args, **kwargs):
        pass


import tensorflow as tf

class qAUC(tf.keras.metrics.AUC):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred_ = (y_pred+1)/2
        super().update_state(y_true, y_pred_, sample_weight=sample_weight)


from abc import ABC, abstractmethod
from inspect import signature, isclass
from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                    List, Optional, overload, Sequence, Set, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)
import sympy as sp
import numpy as np

import cirq
from cirq.ops.gate_features import SingleQubitGate, TwoQubitGate

from quple import QuantumCircuit, TemplateCircuitBlock
from quple.components.interaction_graphs import interaction_graph
from quple.utils.utils import merge_pqc, get_unique_symbols


class ParameterisedCircuit(QuantumCircuit, ABC):
    """Parameterised Quantum Circuit (PQC)
    
    The `ParameterisedCircuit` architecture consists of alternating rotation and entanglement 
    layers that are repeated for a certain number of times. In both layers, parameterized 
    circuit-blocks act on the circuit in a defined way. The rotation layer consists of single 
    qubit gate operations (rotation blocks) that are applied to every qubit in the circuit. 
    The entanglement layer consists of two (or multiple) qubit gate operations (entanglement blocks) 
    applied to the set of qubits defined by an interaction graph.
    
    Common single qubit gates in the rotation layer include the Hadamard gate,
    Pauli X, Pauli Y, Pauli Z gate and the corresponding rotation gates
    
    Common two qubit gates in the entangling layer include the CNOT (CX) gate,
    CZ gate, XX, YY, ZZ gate and the corresponding rotation gates.
    

    Examples:
        >> cq = ParameterisedCircuit(n_qubit=4)
        >> cq.build(rotation_blocks=['H','RZ'],
                    entanglement_blocks=['CNOT'],
                    entangle_strategy='linear',
                    copies=2)
        >> cq
        (0, 0): ───H───Rz(θ_0)───@───────H───Rz(θ_4)───@─────────────────
                                 │                     │
        (0, 1): ───H───Rz(θ_1)───X───@───H───Rz(θ_5)───X─────────@───────
                                     │                           │
        (0, 2): ───H───Rz(θ_2)───────X───@───H─────────Rz(θ_6)───X───@───
                                         │                           │
        (0, 3): ───H───Rz(θ_3)───────────X───H─────────Rz(θ_7)───────X───
    Args:
        n_qubit: int
           Number of qubits in the circuit
        copies: int
           Number of times the layers are repeated.
        rotation_blocks: A list of single qubit gate operations to be applied in the rotation layer.
        entanglement_blocks: A list of multi qubit gate operations to be applied in the entanglement layer.
        entangle_strategy: default='full'
            Determines how the qubits are connected in an entanglement block.
            If str, it specifies the name of the strategy.
            If callable, it specifies the function to map to an interaction graph.
            If list of str, it specifies the names of a list of strategies. The 
            strategy to use is decided by the current block index. For example, if
            the circuit is building the n-th entanglement block in the entanglement layer,
            then the n-th strategy in the list will be used.
            If list of callable, it specifies the list of functions to map to an interaction
            graph. The function to use is decided by the current block index.
        parameter_symbol: str
           Symbol prefix for circuit parameters.
        final_rotation_layer: boolean, default=False
           Whether to add an extra final rotation layer to the circuit.
        flatten_circuit: boolean, default=False
           Whether to flatten circuit parameters when the circuit is modified.
        name: str
           Name of the circuit.    
    """
    def __init__(self, n_qubit:int, copies:int=1,
                rotation_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                parameter_symbol:str='θ',
                final_rotation_layer:bool=False,
                flatten_circuit:bool=False,
                reuse_param_per_depth:bool=False,
                reuse_param_per_layer:bool=False,
                reuse_param_per_template:bool=False,
                parameter_index:Optional[int]=None,
                parameter_scale=1,
                name:str='ParameterisedCircuit',
                *args, **kwargs):
        """ Creates a parameterised circuit following the scheme of 
            alternating rotation and entanglement layers 
           
            Args:
                n_qubit: int
                   Number of qubits in the circuit
                copies: int
                   Number of times the layers are repeated.
                rotation_blocks: A list of single qubit gate operations to be applied in the rotation layer.
                entanglement_blocks: A list of multi qubit gate operations to be applied in the entanglement layer.
                entangle_strategy: default=None
                    Determines how the qubits are connected in an entanglement block.
                    If None, it defaults to using full entanglement.
                    If str, it specifies the name of the strategy.
                    If callable, it specifies the function to map to an interaction graph.
                    If list of str, it specifies the names of a list of strategies. The 
                    strategy to use is decided by the current block index. For example, if
                    the circuit is building the n-th entanglement block in the entanglement layer,
                    then the n-th strategy in the list will be used.
                    If list of callable, it specifies the list of functions to map to an interaction
                    graph. The function to use is decided by the current block index.
                parameter_symbol: str
                   Symbol prefix for circuit parameters.
                final_rotation_layer: boolean, default=False
                   Whether to add an extra final rotation layer to the circuit.
                flatten_circuit: boolean, default=False
                   Whether to flatten circuit parameters when the circuit is modified.
                reuse_param_per_depth: boolean, default=False
                   Whether to reuse parameter symbols at every new depth (symbol starting index reset to 0)
                reuse_param_per_layer: boolean, default=False
                   Whether to reuse parameter symbols at every new layer (symbol starting index reset to 0)
                reuse_param_per_template: boolean, default=False
                   Whether to reuse parameter symbols at every new template block (symbol starting index reset to 0)
                parameter_index: int, default=None
                   Starting index of the first parameter
                name: str
                   Name of the circuit. 
        """
        super().__init__(n_qubit, name=name, *args, **kwargs)
        self._parameter_symbol = parameter_symbol
        self._parameters = np.array([], dtype=object)
        self._readout_qubit = None
        self._flatten_circuit = flatten_circuit
        self._entangle_strategy = entangle_strategy if entangle_strategy else 'full'
        self._parameter_index = parameter_index
        self._reuse_param_per_depth    = reuse_param_per_depth
        self._reuse_param_per_layer    = reuse_param_per_layer
        self._reuse_param_per_template = reuse_param_per_template
        self._parameter_scale = parameter_scale        
        self.build(rotation_blocks, entanglement_blocks, entangle_strategy, copies,
                  final_rotation_layer)
    
    @property
    def entangle_strategy(self):
        return self._entangle_strategy
        
    @property
    def readout_qubit(self):
        """Returns the readout qubit of the circuit
        """
        return self._readout_qubit
    
    @property
    def parameter_index(self):
        """Returns the starting index for the next new parameter symbol
        """
        return self._parameter_index
    
    @property
    def num_param(self):
        """Number of parameters in the circuit created through the 
           add rotation layer and add entanglement layer methods
        """
        return len(self._parameters)
    
    @property
    def parameter_scale(self):
        return self._parameter_scale
    
    def new_param(self, size:int=1, as_array=False) -> sp.Symbol:
        """Returns a new parameter(s) with updated array index(indices)
        
        Args:
            size: int
                Number of new parameters to return 
        
        Returns:
            New parameter(s) with updated array index(indices)
        """
        if size < 1:
            raise ValueError('Number of new parameters created must be greater than zero')
            
        if self.parameter_index is None:
            start = self.num_param
        else:
            start = self.parameter_index
            self._parameter_index += size
            
        params = np.array([sp.Symbol('%s_%s' % (self.parameter_symbol, i)) \
                          for i in range(start, start+size)])
        all_params = np.append(self._parameters, params)
        self._parameters = np.array(get_unique_symbols(all_params))
        
        if (len(params) == 1) and (not as_array):
            return params[0]*self.parameter_scale

        return params*self.parameter_scale
    
    def reset_index(self, index:int=0):
        self._parameter_index = index        
        
    @property
    def parameters(self):
        """Returns an array of parameters in the circuit
        """
        return self._parameters
    
    @property
    def flatten_circuit(self):
        """Whether to flatten the circuit upon construction or modification
        """
        return self._flatten_circuit
    
    @property
    def parameter_symbol(self) -> str:
        """The symbol prefix for the arrays of parameters in the circuit
        """
        return self._parameter_symbol

    def _parse_rotation_blocks(self, rotation_blocks):
        result = ParameterisedCircuit._parse_blocks(rotation_blocks)
        ParameterisedCircuit._validate_blocks(result, SingleQubitGate)
        return result
    
    def _parse_entanglement_blocks(self, entanglement_blocks):
        result = ParameterisedCircuit._parse_blocks(entanglement_blocks)
        ParameterisedCircuit._validate_blocks(result, TwoQubitGate)
        return result
        
    @staticmethod
    def _parse_blocks(blocks) -> List:
        if not blocks:
            return []
        blocks = [blocks] if not isinstance(blocks, list) else blocks
        blocks = [QuantumCircuit._parse_gate_operation(block) \
                  if not isinstance(block, TemplateCircuitBlock) \
                  else block for block in blocks]
        return blocks
    
    @staticmethod
    def _validate_blocks(blocks, gate_feature=None) -> None:
        for block in blocks:
            if isinstance(block, TemplateCircuitBlock):
                continue
            elif isinstance(block, cirq.Gate):
                if gate_feature and \
                    not isinstance(block, gate_feature):
                    raise ValueError('Gate operation {} should be a subclass of '
                                     '{}'.format(type(block), gate_feature))
            elif isclass(block) and issubclass(block, cirq.Gate):
                parameters = ParameterisedCircuit._get_gate_parameters(block)
                n_param = len(parameters)
                dummpy = sp.symarray('x',n_param)
                kwargs = {parameters[i]:dummpy[i] for i in range(n_param)}
                ParameterisedCircuit._validate_blocks([block(**kwargs)], gate_feature)
            elif callable(block):
                n_param = ParameterisedCircuit._get_parameter_count(block)
                if not n_param:
                    raise ValueError('not a valid gate operation or'
                                     'circuit block: {}'.format(block))
                dummy_params = tuple(sp.symarray('x',n_param))
                ParameterisedCircuit._validate_blocks([block(*dummy_params)], gate_feature)
    @staticmethod
    def _get_gate_parameters(gate):
        parameters = []
        parameters_dict = dict(signature(gate).parameters)
        for param_name in parameters_dict:
            annotation = parameters_dict[param_name].annotation
            if (annotation == sp.Basic) or (hasattr(annotation, '__args__') and sp.Basic in annotation.__args__):
                parameters.append(param_name)
        return parameters
      
    @staticmethod
    def _get_parameter_count(gate_expr) -> int:
        '''Returns the number of parameters required for a gate operation expression
        a gate operation expression can be a cirq.Gate instance, a string that can be mapped
        to a cirq.Gate instance, a function that maps to cirq.Gate or a TemplateCircuitBlock instance
        '''
        if isinstance(gate_expr, cirq.Gate):
            return 0
        elif isclass(gate_expr) and issubclass(gate_expr, cirq.Gate):
            return len(ParameterisedCircuit._get_gate_parameters(gate_expr))
        elif isinstance(gate_expr, TemplateCircuitBlock):
            return gate_expr.num_params
        elif callable(gate_expr):
            return len(signature(gate_expr).parameters)
        else:
            raise ValueError('invalid gate expression {} of type {}'.format(gate_expr, type(gate_expr)))

    def get_wavefunction(self, vals:np.ndarray):
        """Returns the simulated wavefunction of the circuit
        """
        simulator = cirq.Simulator()
        param_resolver = self.get_parameter_resolver(vals)
        return simulator.simulate(self, param_resolver=param_resolver).final_state   
    
    
    def add_readout(self, gate:Union[str, cirq.Gate]='XX',
                    readout_qubit:Optional['cirq.GridQubit']=None):
        """Adds a readout qubit to the parameterised circuit
        
        The readout qubit will be entangled to all input qubits via certain two qubit gate operations.
        
        Args:
            gate: The gate operation for entangling the readout qubit to input qubits. 
            readout_qubit: The qubit as readout. Default as (-1, -1).
        """
        parsed_gate = QuantumCircuit._parse_gate_operation(gate)
        readout_qubit = cirq.GridQubit(-1, -1) if readout_qubit is None else readout_qubit
        for qubit in self.qubits:
            final_gate = self.parameterise_gate(parsed_gate)
            self.apply_gate_operation(final_gate, (readout_qubit, qubit))
        if self.flatten_circuit:
            self.flatten()            
        self._readout_qubit = readout_qubit
    
    def readout_measurement(self):
        """Returns the readout gate operation for the readout qubit
        """
        if self.readout_qubit is None:
            raise ValueError('no readout qubit defined in the circuit')
        return cirq.Z(self.readout_qubit)
        
    def merge(self, other):
        """Merges the current circuit with another parameterised circuit
        
        This will update the parameters in the current circuit. If the current 
        circuit is flattened, the merged circuit will be flattened as well.
        
        Args:
            other: the parameterised circuit to be merged with the current circuit
        """
        self._moments = merge_pqc([self, other])._moments
        self._parameters = sp.symarray(self.parameter_symbol, len(self.symbols))
        if self.flatten_circuit:
            self.flatten()
                        
    def get_interaction_graphs(self, block_index:int, num_block_qubits:int,
                               entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                               *args, **kwargs):
        """Returns the interaction graphs for an entanglement block
        
        Args:
            block_index: int
                Index of the current entanglement block in the entanglement layer.
            num_block_qubits: int
                Number of qubits in the entanglement block. That is the number of 
                qubits involved in a gate operation of the entanglement block.
            entangle_strategy: Determines how the qubits are entangled.
                If str, it specifies the name of the strategy.
                If callable, it specifies the function to map to an interaction graph.
                If list of str, it specifies the names of a list of strategies. The 
                strategy to use is decided by the current block index. For example, if
                the circuit is building the n-th entanglement block in the entanglement layer,
                then the n-th strategy in the list will be used.
                If list of callable, it specifies the list of functions to map to an interaction
                graph. The function to use is decided by the current block index.
        Returns:
            A list of tuples specifying the interaction graphs for an entanglement block
        """
        strategy = entangle_strategy if entangle_strategy is not None else self.entangle_strategy
        if isinstance(strategy, str):
            return interaction_graph[strategy](self.n_qubit, num_block_qubits)
        elif callable(strategy):
            return strategy(self.n_qubit, num_block_qubits)
        elif isinstance(strategy, list):
            if all(isinstance(strat, str) for strat in strategy):
                return interaction_graph[strategy[block_index]](self.n_qubit, num_block_qubits)
            elif all(callable(strat) for strat in strategy):
                return strategy[block_index](self.n_qubit, num_block_qubits)
            elif all(isinstance(strat, tuple) for strat in strategy):
                return strategy
        else:
            raise ValueError('invalid entangle strategy: {}'.format(strategy))
        
    def parameterise_gate(self, gate:Union[cirq.Gate, Callable]):
        """Parameterises a gate operation
        
        The gate operation will be parameterised by new symbols from the parameterised circuit.
        This will update the parameter arrays in the circuit.
        
        Args:
            gate: The gate operation to be parameterised
        Returns:
            A parameterised gate operation
        """        
        if isinstance(gate, cirq.Gate):
            if isinstance(gate, (cirq.XXPowGate, cirq.YYPowGate, cirq.ZZPowGate)):
                param = self.new_param()
                return gate**param
            return gate
        elif isclass(gate) and issubclass(gate, cirq.Gate):
            parameters = ParameterisedCircuit._get_gate_parameters(gate)
            n_param = len(parameters)
            params = self.new_param(size=n_param, as_array=True)
            kwargs = {parameters[i]:params[i] for i in range(n_param)}
            return gate(**kwargs)    
        else:
            n_param = ParameterisedCircuit._get_parameter_count(gate)
            params = self.new_param(size=n_param)
            params = tuple(params) if isinstance(params, np.ndarray) else (params,)          
            return gate(*params)
        
    def add_rotation_layer(self, rotation_blocks:Optional[Union[str, cirq.Gate, Callable,
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None):
        """Adds a rotation layer to the parameterised circuit
        
        Args:
            rotation_blocks: A list of single qubit gate operations to be applied in the rotation layer.
        """
        rotation_blocks = self._parse_rotation_blocks(rotation_blocks)
        for i, block in enumerate(rotation_blocks):
            if self._reuse_param_per_layer:
                self.reset_index()
            if isinstance(block, TemplateCircuitBlock):
                interaction_graphs = self.get_interaction_graphs(i, block.num_block_qubits)
                for qubits in interaction_graphs:
                    if self._reuse_param_per_template:
                        self.reset_index()
                    block.build(self, qubits)
            else:
                for qubit in range(self.n_qubit):
                    gate = self.parameterise_gate(block)
                    self.apply_gate_operation(gate, qubit)
                
    def add_entanglement_layer(self, entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None):
        """Adds an entanglement layer to the parameterised circuit
        
        Args:
            entanglement_blocks: A list of multi qubit gate operations to be applied in the entanglement layer.
            entangle_strategy: Determines how the qubits are entangled in an entanglement block.
                If str, it specifies the name of the strategy.
                If callable, it specifies the function to map to an interaction graph.
                If list of str, it specifies the names of a list of strategies. The 
                strategy to use is decided by the current block index. For example, if
                the circuit is building the n-th entanglement block in the entanglement layer,
                then the n-th strategy in the list will be used.
                If list of callable, it specifies the list of functions to map to an interaction
                graph. The function to use is decided by the current block index.
        """       
        entangle_strategy = entangle_strategy or 'full' 
        entanglement_blocks = self._parse_entanglement_blocks(entanglement_blocks)
        for i, block in enumerate(entanglement_blocks):
            if self._reuse_param_per_layer:
                self.reset_index()            
            if isinstance(block, TemplateCircuitBlock):
                interaction_graphs = self.get_interaction_graphs(i, block.num_block_qubits,
                                                                entangle_strategy)
                for qubits in interaction_graphs:
                    if self._reuse_param_per_template:
                        self.reset_index()                    
                    block.build(self, qubits)
            else:
                interaction_graphs = self.get_interaction_graphs(i, 2, entangle_strategy)
                for qubits in interaction_graphs:
                    gate = self.parameterise_gate(block)
                    self.apply_gate_operation(gate, qubits)
                    
                    
    def build(self, rotation_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                copies:int=1, final_rotation_layer:bool=False):
        """Builds a parameterised circuit following the scheme of 
           alternating rotation and entanglement layers 
           
           Args:
               rotation_blocks: A list of single qubit gate operations to be applied in the rotation layer.
               entanglement_blocks: A list of multi qubit gate operations to be applied in the entanglement layer.
               entangle_strategy: Determines how the qubits are entangled in an entanglement block.
                   If str, it specifies the name of the strategy.
                   If callable, it specifies the function to map to an interaction graph.
                   If list of str, it specifies the names of a list of strategies. The 
                   strategy to use is decided by the current block index. For example, if
                   the circuit is building the n-th entanglement block in the entanglement layer,
                   then the n-th strategy in the list will be used.
                   If list of callable, it specifies the list of functions to map to an interaction
                   graph. The function to use is decided by the current block index.
               copies: int
                   Number of times the layers are repeated.
                   
            Examples:
                >> cq = ParameterisedCircuit(n_qubit=4)
                >> cq.build(rotation_blocks=['H','RZ'],
                            entanglement_blocks=['CNOT'],
                            entangle_strategy='linear',
                            copies=2)
                >> cq
                (0, 0): ───H───Rz(θ_0)───@───────H───Rz(θ_4)───@─────────────────
                                         │                     │
                (0, 1): ───H───Rz(θ_1)───X───@───H───Rz(θ_5)───X─────────@───────
                                             │                           │
                (0, 2): ───H───Rz(θ_2)───────X───@───H─────────Rz(θ_6)───X───@───
                                                 │                           │
                (0, 3): ───H───Rz(θ_3)───────────X───H─────────Rz(θ_7)───────X───
        """
        self.clear()
        for _ in range(copies):
            if self._reuse_param_per_depth:
                self.reset_index()
            self.add_rotation_layer(rotation_blocks)
            self.add_entanglement_layer(entanglement_blocks, entangle_strategy)
            
        if final_rotation_layer:
            self.add_rotation_layer(rotation_blocks)
            
        if self.flatten_circuit:
            self.flatten()                    
    
    def clear(self) -> None:
        """Clears all gate operations in the circuit
        """
        super().clear()
        self._parameters = np.array([], dtype=object)
        
        
    def run_simulator(self, repetitions:int=1):
        pass
        #simulator = cirq.Simulator()
        #simulator.run(resolved_circuit, repetitions=repetitions)
        ## to get wave function:
        #simulator.simulate(resolved_circuit)
        #cirq.measure_state_vector   
        #output_state_vector = simulator.simulate(self, resolver).final_state
        #z0 = cirq.X(q0)
        #qubit_map = {q0: 0, q1: 1}
        #z0.expectation_from_wavefunction(output_state_vector, qubit_map).real


import numbers
import numpy as np
import tensorflow as tf

import cirq
import sympy
from tensorflow_quantum.python.layers.circuit_executors import \
    expectation, sampled_expectation
from tensorflow_quantum.python.layers.circuit_construction import elementary
from tensorflow_quantum.python import util

import quple
from quple import QuantumCircuit
from quple.interface.tfq.tf_resolvers import resolve_formulas

class PQC(tf.keras.layers.Layer):
    """Parametrized Quantum Circuit (PQC) Layer.
    This layer is for training parameterized quantum models.
    Given a parameterized circuit, this layer initializes the parameters
    and manages them in a Keras native way.
    We start by defining a simple quantum circuit on one qubit.
    This circuit parameterizes an arbitrary rotation on the Bloch sphere in
    terms of the three angles a, b, and c:
    >>> q = cirq.GridQubit(0, 0)
    >>> (a, b, c) = sympy.symbols("a b c")
    >>> circuit = cirq.Circuit(
    ...     cirq.rz(a)(q),
    ...     cirq.rx(b)(q),
    ...     cirq.rz(c)(q),
    ...     cirq.rx(-b)(q),
    ...     cirq.rz(-a)(q)
    ... )
    In order to extract information from our circuit, we must apply measurement
    operators.  For now we choose to make a Z measurement.  In order to observe
    an output, we must also feed our model quantum data (NOTE: quantum data
    means quantum circuits with no free parameters).  Though the output values
    will depend on the default random initialization of the angles in our model,
    one will be the negative of the other since `cirq.X(q)` causes a bit flip:
    >>> outputs = tfq.layers.PQC(circuit, cirq.Z(q))
    >>> quantum_data = tfq.convert_to_tensor([
    ...     cirq.Circuit(),
    ...     cirq.Circuit(cirq.X(q))
    ... ])
    >>> res = outputs(quantum_data)
    >>> res
    <tf.Tensor: id=577, shape=(2, 1), dtype=float32, numpy=
    array([[ 0.8722095],
           [-0.8722095]], dtype=float32)>
    We can also choose to measure the three pauli matrices, sufficient to
    fully characterize the operation of our model, or choose to simulate
    sampled expectation values by specifying a number of measurement shots
    (repetitions) to average over.  Notice that using only 200 repetitions
    introduces variation between the two rows of data, due to the
    probabilistic nature of measurement.
    >>> measurement = [cirq.X(q), cirq.Y(q), cirq.Z(q)]
    >>> outputs = tfq.layers.PQC(circuit, measurement, repetitions=200)
    >>> quantum_data = tfq.convert_to_tensor([
    ...     cirq.Circuit(),
    ...     cirq.Circuit(cirq.X(q))
    ... ])
    >>> res = outputs(quantum_data)
    >>> res
    <tf.Tensor: id=808, shape=(2, 3), dtype=float32, numpy=
    array([[-0.38,  0.9 ,  0.14],
           [ 0.19, -0.95, -0.35]], dtype=float32)>
    A value for `backend` can also be supplied in the layer constructor
    arguments to indicate which supported backend you would like to use.
    A value for `differentiator` can also be supplied in the constructor
    to indicate the differentiation scheme this `PQC` layer should use.
    Here's how you would take the gradients of the above example using a
    `cirq.Simulator` backend (which is slower than the default
    `backend=None` which uses C++):
    >>> q = cirq.GridQubit(0, 0)
    >>> (a, b, c) = sympy.symbols("a b c")
    >>> circuit = cirq.Circuit(
    ...     cirq.rz(a)(q),
    ...     cirq.rx(b)(q),
    ...     cirq.rz(c)(q),
    ...     cirq.rx(-b)(q),
    ...     cirq.rz(-a)(q)
    ... )
    >>> measurement = [cirq.X(q), cirq.Y(q), cirq.Z(q)]
    >>> outputs = tfq.layers.PQC(
    ...     circuit,
    ...     measurement,
    ...     repetitions=5000,
    ...     backend=cirq.Simulator(),
    ...     differentiator=tfq.differentiators.ParameterShift())
    >>> quantum_data = tfq.convert_to_tensor([
    ...     cirq.Circuit(),
    ...     cirq.Circuit(cirq.X(q))
    ... ])
    >>> res = outputs(quantum_data)
    >>> res
    <tf.Tensor: id=891, shape=(2, 3), dtype=float32, numpy=
    array([[-0.5956, -0.2152,  0.7756],
           [ 0.5728,  0.1944, -0.7848]], dtype=float32)>
    Lastly, like all layers in TensorFlow the `PQC` layer can be called on any
    `tf.Tensor` as long as it is the right shape. This means you could replace
    `quantum_data` with values fed in from a `tf.keras.Input`.
    """

    def __init__(
            self,
            model_circuit,
            data_circuit,
            operators,
            *,
            repetitions=None,
            backend=None,
            differentiator=None,
            initializer=None,
            regularizer=None,
            constraint=None,
            trainable=True,
            seed=None,
            name=None,
            **kwargs,
    ):
        """Instantiate this layer.
        Create a layer that will output expectation values of the given
        operators when fed classical data to it's input layer. This layer will
        accept one input tensor representing a classical data source which is used
        to resolve the values of the parameter expressions in data_circuit. The resolved
        data_circuit will be appended to the model_circuit which is then executed and
        finally output the expectation values.
        
        Arguments:
            model_circuit: `cirq.Circuit` or `quple.QuantumCircuit` containing 
                `sympy.Symbols` that will be used as the model which will be fed with 
                quantum data from the data circuit.
            data_circuit: `cirq.Circuit` or `quple.QuantumCircuit` containing 
                `sympy.Symbols` that maps the inputs in this layer to the quantum
                data of the circuit which will be fed to the model circuit.
            operators: `cirq.PauliSum` or Python `list` of `cirq.PauliSum` objects
                used as observables at the end of the model circuit.
            repetitions: Optional python `int` indicating how many samples to use
                when estimating expectation values.  If `None` analytic expectation
                calculation is used.
            backend: Optional backend to use to simulate states. Defaults to
                the native TensorFlow simulator (None), however users may also
                specify a preconfigured cirq simulation object to use instead.
                If a cirq object is given it must inherit either
                `cirq.SimulatesFinalState` if analytic expectations are desired or
                `cirq.Sampler` if sampled expectations are desired.
            differentiator: Optional `tfq.differentiator` object to specify how
                gradients of `model_circuit` should be calculated.
            initializer: Optional `tf.keras.initializer` object to specify how the
                symbols in `model_circuit` should be initialized when creating
                the managed variables. If None, defaults to initialize from a
                random uniform distribution between 0 and 2*pi. 
            regularizer: Optional `tf.keras.regularizer` object applied to the
                managed variables parameterizing `model_circuit`.
            constraint: Optional `tf.keras.constraint` object applied to the
                managed variables parameterizing `model_circuit`.
            trainable: Boolean, whether the layer's variables should be trainable.
            seed: Optional integer. Controls the random seed of the weight initializer.
            name: Optional string. Name of the layer.
        """
        super().__init__(trainable=trainable, name=name, **kwargs)

        # Ingest model_circuit.
        if not isinstance(model_circuit, cirq.Circuit):
            raise TypeError("model_circuit must be a cirq.Circuit object."
                            " Given: {}".format(model_circuit))
        # Ingest data_circuit.
        if not isinstance(data_circuit, cirq.Circuit):
            raise TypeError("data_circuit must be a cirq.Circuit object."
                            " Given: {}".format(data_circuit))
            
        if not isinstance(data_circuit, QuantumCircuit):
            data_circuit = QuantumCircuit.from_cirq(data_circuit)
        if not isinstance(model_circuit, QuantumCircuit):
            model_circuit = QuantumCircuit.from_cirq(model_circuit)
            
        if quple.has_composite_symbols(model_circuit):
            raise ValueError("model_circuit symbols must not be composite")            
        
        data_circuit.flatten()
        data_circuit_qubits = data_circuit.qubits
        model_circuit_qubits = model_circuit.qubits
        unaccounted_qubits = set(data_circuit_qubits) - set(model_circuit_qubits)
        if len(unaccounted_qubits) != 0:
            raise ValueError("model_circuit missing the following input qubits from"
                             " the data_circuit: ".format(unaccounted_qubits))
        self._data_qubits = data_circuit_qubits
        self._n_qubit = len(data_circuit_qubits)
        self._input_symbols_list = [v.name for v in data_circuit.expr_map.values()]
        self._base_input_symbols_list = data_circuit.raw_symbols
        self._data_circuit_formulas = list(data_circuit.expr_map)
        self._num_formulas = tf.constant(len(self._data_circuit_formulas))
        self._symbols_list = model_circuit.symbols
        
        if not set(self._base_input_symbols_list).isdisjoint(set(self._symbols_list)):
            raise ValueError("data_circuit and model circuit must not share symbols of the same name")

        self._input_symbols = tf.constant(self._input_symbols_list)
        self._symbols = tf.constant(self._symbols_list)

        self._all_symbols_list = self._input_symbols_list + self._symbols_list
        self._all_symbols = tf.constant(self._all_symbols_list)

        self._model_circuit = util.convert_to_tensor([model_circuit])
        self._data_circuit = util.convert_to_tensor([data_circuit])
        if len(self._symbols_list) == 0:
            raise ValueError("model_circuit has no sympy.Symbols. Please "
                             "provide a circuit that contains symbols so "
                             "that their values can be trained.")
            
        if len(self._input_symbols_list) == 0:
            raise ValueError("data_circuit has no sympy.Symbols. Please "
                             "provide a circuit that contains symbols so "
                             "that their values can be passed as input.")            

        # Ingest operators.
        if isinstance(operators, (cirq.PauliString, cirq.PauliSum)):
            operators = [operators]
        if not isinstance(operators, (list, np.ndarray, tuple)):
            raise TypeError("operators must be a cirq.PauliSum or "
                            "cirq.PauliString, or a list, tuple, "
                            "or np.array containing them. "
                            "Got {}.".format(type(operators)))
        if not all([
                isinstance(op, (cirq.PauliString, cirq.PauliSum))
                for op in operators
        ]):
            raise TypeError("Each element in operators to measure "
                            "must be a cirq.PauliString"
                            " or cirq.PauliSum")
        self._operators = util.convert_to_tensor([operators])

        # Ingest and promote repetitions.
        self._analytic = False
        if repetitions is None:
            self._analytic = True
        if not self._analytic and not isinstance(repetitions, numbers.Integral):
            raise TypeError("repetitions must be a positive integer value."
                            " Given: ".format(repetitions))
        if not self._analytic and repetitions <= 0:
            raise ValueError("Repetitions must be greater than zero.")
        if not self._analytic:
            self._repetitions = tf.constant(
                [[repetitions for _ in range(len(operators))]],
                dtype=tf.dtypes.int32)

        # Set backend and differentiator.
        if not isinstance(backend, cirq.Sampler
                         ) and repetitions is not None and backend is not None:
            raise TypeError("provided backend does not inherit cirq.Sampler "
                            "and repetitions!=None. Please provide a backend "
                            "that inherits cirq.Sampler or set "
                            "repetitions=None.")
        if not isinstance(backend, cirq.SimulatesFinalState
                         ) and repetitions is None and backend is not None:
            raise TypeError("provided backend does not inherit "
                            "cirq.SimulatesFinalState and repetitions=None. "
                            "Please provide a backend that inherits "
                            "cirq.SimulatesFinalState or choose a positive "
                            "number of repetitions.")
        if self._analytic:
            self._executor = expectation.Expectation(
                backend=backend, differentiator=differentiator)
        else:
            self._executor = sampled_expectation.SampledExpectation(
                backend=backend, differentiator=differentiator)

        self._append_layer = elementary.AddCircuit()

        # Set additional parameter controls.
        if initializer is None:
            initializer = tf.keras.initializers.RandomUniform(0, 2 * np.pi, seed=seed)
        self.initializer = tf.keras.initializers.get(initializer)
        self.regularizer = tf.keras.regularizers.get(regularizer)
        self.constraint = tf.keras.constraints.get(constraint)

        self._init_weights()
        self._validate_init()
        
    def _validate_init(self):
        pass
    
    def _validate_build(self, input_shape):
        num_input_symbols = len(self._base_input_symbols_list)
        sum_input_shape = tf.reduce_sum(input_shape[1:])
        if num_input_symbols != sum_input_shape:
            raise ValueError(f"sum of input shape (={sum_input_shape}, excluding batch dimension) must match the"
                             f" number of input symbols in data circuit (={num_input_symbols})")
        
    def _get_input_resolver(self):
        return resolve_formulas(self._data_circuit_formulas,
                                self._base_input_symbols_list)
        
    def _init_weights(self):
        """Initialize trainable weights.
        """
        # Weight creation is not placed in a build function because the number
        # of weights is independent of the input shape.
        self.parameters = self.add_weight('parameters',
                                          shape=self._symbols.shape,
                                          initializer=self.initializer,
                                          regularizer=self.regularizer,
                                          constraint=self.constraint,
                                          dtype=tf.float32,
                                          trainable=True)

    @property
    def symbols(self):
        """The model symbols that are managed by this layer (in-order).
        
        Note: `symbols[i]` indicates what symbol name the managed weights corresponding to 
        the model circuit map to.
        """
        return [sympy.Symbol(x) for x in self._symbols_list]

    def symbol_values(self):
        """Returns a Python `dict` containing the model symbol name, value pairs.
        Returns:
            Python `dict` with `str` keys and `float` values representing
                the current symbol values.
        """
        return dict(zip(self.symbols, self.get_weights()[0]))
    
    @property
    def input_symbols(self):
        """The flattened input symbols that are managed by this layer (in-order).
        """
        return [sympy.Symbol(x) for x in self._input_symbols_list]
    
    @property
    def base_input_symbols(self):
        """The base (non-flattened) input symbols that are managed by this layer (in-order).
        
        Note: `symbols[i]` indicates what symbol name the inputs in this layer map to.
        """
        return [sympy.Symbol(x) for x in self._base_input_symbols_list]     
    
    @property
    def all_symbols(self):
        """The flattened input and model symbols that are managed by this layer (in-order).
        """
        return [sympy.Symbol(x) for x in self._all_symbols_list]     

    def build(self, input_shape):
        """Keras build function."""
        self._input_resolver = self._get_input_resolver()
        self._validate_build(input_shape)
        self.built = True

    def call(self, inputs):
        """Keras call function."""
        circuit_batch_dim = tf.gather(tf.shape(inputs), 0)
        tiled_up_data_circuit = tf.tile(self._data_circuit, [circuit_batch_dim])
        tiled_up_model = tf.tile(self._model_circuit, [circuit_batch_dim])
        model_appended = self._append_layer(tiled_up_data_circuit, append=tiled_up_model)
        tiled_up_parameters_ = tf.tile([self.parameters], [circuit_batch_dim, 1])
        resolved_inputs = self._input_resolver(inputs)     
        tiled_up_parameters = tf.concat([resolved_inputs, tiled_up_parameters_], 1)
        tiled_up_operators = tf.tile(self._operators, [circuit_batch_dim, 1])

        # this is disabled to make autograph compilation easier.
        # pylint: disable=no-else-return
        if self._analytic:
            return self._executor(model_appended,
                                  symbol_names=self._all_symbols,
                                  symbol_values=tiled_up_parameters,
                                  operators=tiled_up_operators)
        else:
            tiled_up_repetitions = tf.tile(self._repetitions,
                                           [circuit_batch_dim, 1])
            return self._executor(model_appended,
                                  symbol_names=self._all_symbols,
                                  symbol_values=tiled_up_parameters,
                                  operators=tiled_up_operators,
                                  repetitions=tiled_up_repetitions)
        # pylint: enable=no-else-return


from typing import List, Union, Optional, Callable, Sequence, Tuple

from quple import ParameterisedCircuit

SU2_gate_set = ['RX','RY','RZ','X','Y','Z']

class EfficientSU2(ParameterisedCircuit):
    '''The efficient SU2 parameterised circuit
    
    Implementation based on the Qiskit library.
    
    The EfficientSU2 circuit consists of layers of single qubit
    operations spanned by SU(2) abd a layer of CNOT entanglements.
    This construction is believed to be hardware efficient.
    
    The EfficientSU2 circuit may be used as a variational circuit
    or a model circuit in the PQC layer of a machine learning model.
    It represents a unitary U(θ) parameterised by a set of free 
    parameters θ. The circuit may be repeated several times to 
    increase the number of parameters and hence the complexity 
    of the model. 
    
    It is an example of a strongly entangling circuit which has 
    the advantage of capturing correlations between the data
    features at all ranges for a short range circuit. 
    Reference: https://arxiv.org/pdf/1804.00633.pdf
    
    Args:
      n_qubit: int
          Number of qubits in the circuit
      copies: int
         Number of times the layers are repeated.
      su2_gates: 
          The SU2 gates to be used in the circuit.
          If str, it is the name of the SU2 gate operation.
          If list of str, it is the list of names of the SU2 gate operations 
          applied in the given order to every qubit.
          If cirq.Gate type or list of cirq.Gate type, it is the SU2 gate operation
          or list of SU2 gate operations applied in the given order to every qubit.
      entangle_strategy: Determines how the qubits are connected in an entanglement block.
         If str, it specifies the name of the strategy.
         If callable, it specifies the function to map to an interaction graph.
         If list of str, it specifies the names of a list of strategies. The 
         strategy to use is decided by the current block index. For example, if
         the circuit is building the n-th entanglement block in the entanglement layer,
         then the n-th strategy in the list will be used.
         If list of callable, it specifies the list of functions to map to an interaction
         graph. The function to use is decided by the current block index.
      parameter_symbol: str
         Symbol prefix for circuit parameters.
      name: str
         Name of the circuit.    
    '''    
    def __init__(self, n_qubit: int, copies: int=2, 
                 su2_gates:Optional[Union[str, 'cirq.Gate', Callable, 
                                          List[str],List['cirq.Gate'],List[Callable]]]=None,
                 entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                 parameter_symbol:str='θ', name:str='EfficientSU2', *args, **kwargs):
        
        su2_gates = su2_gates or ['RY','RZ']
        super().__init__(n_qubit=n_qubit, copies=copies,
                         rotation_blocks=su2_gates,
                         entanglement_blocks='CX',
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         name=name,
                         flatten_circuit=False,
                         final_rotation_layer=True,
                         *args, **kwargs)


import os
import time
import datetime
import logging

import numpy as np
import quple

from sklearn.metrics import accuracy_score,roc_auc_score

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class DefaultDict(dict):
    def __missing__(self, key):
        return key
    
class QSVMLogger():
    """Logger for the quantum support vector machine classifer (QSVM)
    
    """
    def __init__(self, 
                 log_dir='./logs', 
                 filename='qsvm_'
                 'circuit_{circuit_name}_'
                 'encoder_{encoding_map}_'
                 'entanglement_{entangle_strategy}_'
                 'nqubit_{n_qubit}_'
                 'train_{train_size}_'
                 'test_{test_size}_{time}',
                 keys=None,
                 stream_level=logging.INFO,
                 file_level=logging.DEBUG,
                 formatter=None,
                 save_npz=True,
                 save_weights=True,
                 roc_plot=True,
                 extra_attributes={}):
        """Creates the VQC logger
        Args:
            log_dir: str
                Name of logging directory.
            filename: str
                File name prefix for logged items.
            keys: str
                Keys of epoch results to save.
            stream_level: str
                Stream level for the logger.
            file_level: str
                File level for the logging file.
            formatter: str
                Formatter for logging.
            save_npz: boolean; default=True
                Whether to save VQC specs, epoch results and test results
                to a npz file.
            save_weights: boolean; default=True
                Whether to save the weights of the model.
            roc_plot: boolean; default=True
                Whether to save a roc plot for the test result.
            extra_attributes: dict, defaul={}
                Extra attributes to be saved in the logger
        """
        
        self.log_dir = log_dir
        self.filename = filename
        self.keys = keys
        self.stream_level = stream_level
        self.file_level = file_level
        self.stream_handler = None
        self.file_handler = None
        self.formatter = formatter or logging.Formatter('%(asctime)s [%(threadName)-12.12s]'
                                                         '[%(levelname)-5.5s]  %(message)s')
        self.logger_is_set = False
        self.save_npz = save_npz
        self.save_weights = save_weights
        self.roc_plot = roc_plot
        self.attributes = {}
        self.extra_attributes = extra_attributes
        
    def set_attributes(self, attributes):
        self.attributes = {**attributes, **self.extra_attributes}
        
    def reset_logger(self):
        self.file_handler.close()
        logger.removeHandler(self.stream_handler)
        logger.removeHandler(self.file_handler)
        self.logger_is_set = False
        
    def setup_logger(self):
        if self.logger_is_set:
            return None
        
        # setup time attributes
        self.start_time = time.time()
        self.time = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d_%H-%M-%S-%f')  
        
        # setup stream handler
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(self.stream_level)
        stream_handler.setFormatter(self.formatter)
        self.stream_handler = stream_handler        
        
        # setup file name and formats
        self.attributes['time'] = self.time
        self.attributes = DefaultDict(self.attributes)
        self.formatted_filename = os.path.join(self.log_dir, self.filename.format(**self.attributes))
        log_filename = self.formatted_filename + '.log'
        os.makedirs(os.path.dirname(log_filename), exist_ok=True)
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(self.file_level)
        file_handler.setFormatter(self.formatter)
        self.file_handler = file_handler
        
        #setup logger
        logger.addHandler(self.stream_handler)
        logger.addHandler(self.file_handler)
        self.logger_is_set = True
        
    def print_attributes(self):
        attrib = self.attributes.copy()
        logger.info('######## Executing QSVM with the following attributes ########')
        logger.info('Feature Dimension: {}'.format(attrib.pop('feature_dimension','')))            
        logger.info('Number of Qubits: {}'.format(attrib.pop('n_qubit','')))
        logger.info('Encoding Circuit: {}'.format(attrib.pop('encoding_circuit','')))
        logger.info('Encoding Map: {}'.format(attrib.pop('encoding_map','')))                
        logger.info('Train Size: {}'.format(attrib.pop('train_size','')))
        logger.info('Test Size: {}'.format(attrib.pop('test_size','')))
        for k,v in attrib.items():
            logger.info('{}: {}'.format(k.replace('_',' ').title(),v)) 
        logger.info('##############################################################')
        
    def info(self, text):
        logger.info(text)
        
    def on_train_start(self):
        self.setup_logger()
        self.print_attributes()
        
        
    def on_train_end(self, clf, test_kernel_matrix, y_test, val_kernel_matrix=None, y_val=None):
        logger.info('##################### Training Ends ##########################')
        
        score = clf.predict_proba(test_kernel_matrix)[:,1]
        logger.info("best_params: %s" % clf.best_params_)

        means = clf.cv_results_['mean_test_score']
        stds = clf.cv_results_['std_test_score']

        for mean, std, params in zip(means, stds, clf.cv_results_['params']):
            logger.info("%0.3f (+/-%0.03f) for %r" % (mean, std * 2, params))

        predictions = [round(value) for value in score]

        roc_auc =  roc_auc_score(y_test, score)
        accuracy =  accuracy_score(y_test, predictions)

        logger.info("AUC: %s" % roc_auc)
        logger.info("Accuracy: %s" % accuracy)
        
        self.end_time = time.time()
        
        self.attributes['time_taken'] = self.end_time - self.start_time
        
        result = {'score': score,
                  'y_test': y_test,
                  'auc': roc_auc,
                  'accuracy':accuracy}
        
        extras = {}
            
        if (val_kernel_matrix is not None) and (y_val is not None):
            val_score = clf.predict_proba(val_kernel_matrix)[:,1]
            val_auc =  roc_auc_score(y_val, val_score)
            result.update({'val_score': val_score,
                           'y_val': y_val,
                           'val_auc': val_auc}) 
                
        logger.info("Time taken: %s" % self.attributes['time_taken'])
        np.savez(self.formatted_filename +'.npz', 
                 result=result,
                 **self.attributes,
                 **extras)
        
        self.reset_logger()
        return roc_auc

    
    


import os
from typing import Union, Optional, Callable, Dict, Tuple
from functools import partial

import numpy as np
import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Optimizer

from quple.models import AbstractModel

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

try:
    from IPython.display import clear_output
except ImportError:
    clear_output = None

class QGAN(AbstractModel):
    """Quantum Generative Adversarial Network (QGAN)
    """    
    def __init__(self, generator:Model, 
                 discriminator:Model,
                 latent_dim:Optional[Union[int, Tuple]]=None,
                 n_disc:int=1,
                 n_gen:int=1,
                 epochs:int=100, 
                 batch_size:int=32,
                 optimizer:Optional[Union[str, Dict]]=None,
                 optimizer_kwargs:Optional[Dict]=None,
                 name:str='QGAN',
                 random_state:Optional[int]=None,
                 checkpoint_dir:Optional[str]=None,
                 checkpoint_interval:int=10,
                 checkpoint_max_to_keep:Optional[int]=None):
        """ Creates a QGAN model equipped with a generator and a discriminator neural networks.
            The modified Jensen-Shanon divergence is used as the loss function.
        Args:
            generator: tensorflow.keras.Model
                A keras model representing the generator neural network. It can contain
                both classical and quantum layers.
            discriminator: cirq.Circuit or quple.QuantumCircuit instace
                A keras model representing the discriminator neural network. It can contain
                both classical and quantum layers.
            latent_dim: (Optional) int or tuple of int
                Dimensions of the latent space. It specifies the shape of the input noise
                (drawn from a normal distribution) passed to the generator neural network.
            n_disc: (Optional) int, default=1
                Number of discriminator iterations per generator iteration in each epoch.
            n_disc: (Optional) int, default=1
                Number of generatator iterations per discriminator iteration in each epoch.                
            epochs: int, default=100
                Number of epochs for the training.
            batch_size: int, default=10
                Number of training examples used in one iteration. Note that the true batch
                size is batch_size * n_disc so that training samples of size batch_size is
                passed to the discriminator in each discriminator step.
            name: (Optional) string
                Name given to the model.
            random_state: (Optional) int
                The random state used at the beginning of a training
                for reproducible result.
            checkpoint_dir: Optional[string]
                The path to a directory in which to write checkpoints. If None,
                no checkpoint will be saved.
            checkpoint_interval: int, default=10
                Number of epochs between each checkpoint.
            checkpoint_max_to_keep: (Optional) int
                Number of checkpoints to keep. If None, all checkpoints are kept.
        """
        self.set_random_state(random_state)
        
        self.start_epoch = 0
        self.n_disc = n_disc
        self.n_gen  = n_gen
        self.epochs = epochs
        self.batch_size = batch_size

        if not isinstance(generator, Model):
            raise ValueError("generator must be a tf.keras.Model instance")
        self.G = generator
        if not isinstance(discriminator, Model):
            raise ValueError("discriminator must be a tf.keras.Model instance")            
        self.D = discriminator
        
        if latent_dim is None:
            try:
                self.latent_dim = self.G.input_shape[1:]
            except:
                raise RuntimeError("cannot infer input shape from generator")
        else:
            self.latent_dim = tuple(latent_dim)
        self.z_batch_shape = (self.batch_size,) + self.latent_dim            

        self.G_optimizer, self.D_optimizer = self._get_optimizers(optimizer, optimizer_kwargs)
        
        self.loss_function = self._get_loss_function()
        
        self.visualization_interval = None
        self.image_shape = None
        self.n_image_to_show = 0
        
        print('Summary of Generator')
        self.G.summary()
        
        print('Summary of Discriminator')
        self.D.summary()
        
        super().__init__(name=name, random_state=random_state,
                         checkpoint_dir=checkpoint_dir,
                         checkpoint_interval=checkpoint_interval,
                         checkpoint_max_to_keep=checkpoint_max_to_keep)

    def _create_checkpoint(self):
        checkpoint = tf.train.Checkpoint(step=tf.Variable(1),
                                         generator_optimizer=self.G_optimizer,
                                         discriminator_optimizer=self.D_optimizer,
                                         generator=self.G,
                                         discriminator=self.D)
        return checkpoint
        
        
    def _get_loss_function(self):
        return tf.keras.losses.BinaryCrossentropy(from_logits=True)
        
    @staticmethod
    def _get_optimizers(optimizer:Optional[Union[str, Dict]]=None,
                        optimizer_kwargs:Optional[Dict]=None):
        if isinstance(optimizer, str):
            optimizer = {"generator": optimizer, "discriminator": optimizer}
        elif isinstance(optimizer, dict):
            if ("generator" not in optimizer) or ("discriminator" not in optimizer):
                raise ValueError("optimizer passed as dictionary must contain both  "
                                 "`generator` and `discriminator` keys")
        else:
            raise ValueError("could not interpret optimizer: {}".format(optimizer))
        if optimizer_kwargs is None:
            optimizer_kwargs = {"generator": {}, "discriminator": {}}
        elif isinstance(optimizer_kwargs, dict):
            if ("generator" not in optimizer_kwargs) and ("discriminator" not in optimizer_kwargs):
                optimizer_kwargs = {"generator": optimizer_kwargs, "discriminator": optimizer_kwargs}
            elif ("generator" not in optimizer_kwargs) or ("discriminator" not in optimizer_kwargs):
                raise ValueError("optimizer_kwargs passed as dictionary must contain both  "
                                 "`generator` and `discriminator` keys")
        resolved = {}
        for key in ["generator", "discriminator"]:
            identifier = optimizer[key]
            config = optimizer_kwargs[key]
            if isinstance(identifier, Optimizer):
                resolved[key] = identifier
            elif isinstance(identifier, dict):
                identifier.update(kwargs)
                resolved[key] = tf.keras.optimizers.get(identifier)
            elif isinstance(identifier, str):
                identifier = {"class_name": str(identifier), "config": config}
                resolved[key] = tf.keras.optimizers.get(identifier)
            else:
                raise ValueError("could not interpret identifier: {}".format(identifier))
        generator_optimizer = resolved["generator"]
        discriminator_optimizer = resolved["discriminator"]
        return generator_optimizer, discriminator_optimizer
                        
    
    @tf.function
    def D_loss(self, real_output, fake_output):
        """Compute discriminator loss."""
        real_loss = self.loss_function(tf.ones_like(real_output), real_output)
        fake_loss = self.loss_function(tf.zeros_like(fake_output), fake_output)
        return real_loss + fake_loss
    
    @tf.function
    def G_loss(self, fake_output):
        """Compute generator loss."""
        return self.loss_function(tf.ones_like(fake_output), fake_output)
    
    @tf.function
    def train_step_1v1(self, x_real):
        """Training step for one epoch with 1 generator step and 1 discriminator step
        """
        z = tf.random.normal(self.z_batch_shape)
        with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
            x_fake_ = self.G(z, training=True)
            x_fake = tf.reshape(x_fake_, tf.shape(x_real))
            real_output = self.D(x_real, training=True)
            fake_output = self.D(x_fake, training=True)
            gen_loss = self.G_loss(fake_output)
            disc_loss = self.D_loss(real_output, fake_output)
        grad_gen = gen_tape.gradient(gen_loss, self.G.trainable_variables)
        grad_disc = disc_tape.gradient(disc_loss, self.D.trainable_variables)
        self.G_optimizer.apply_gradients(zip(grad_gen, self.G.trainable_variables))
        self.D_optimizer.apply_gradients(zip(grad_disc, self.D.trainable_variables))  
        return gen_loss, disc_loss
    
    @tf.function 
    def G_step(self):
        """Perform one training step for generator"""
        # using Gaussian noise with mean 0 and width 1 as generator input
        z = tf.random.normal(self.z_batch_shape)
        with tf.GradientTape() as t:
            x_fake = self.G(z, training=True)
            fake_output = self.D(x_fake, training=True)
            loss = self.G_loss(fake_output)
        grad = t.gradient(loss, self.G.trainable_variables)
        self.G_optimizer.apply_gradients(zip(grad, self.G.trainable_variables))
        return loss
    
    @tf.function
    def D_step(self, x_real):
        """Perform one training step for discriminator
        
            Arguments:
                x_real: Numpy array, tf.Tensor / any inputs accepted by tensorflow
                    Input from the real data.
        """
        # using Gaussian noise with mean 0 and width 1 as generator input
        z = tf.random.normal(self.z_batch_shape)
        with tf.GradientTape() as t:
            x_fake_ = self.G(z, training=True)
            x_fake = tf.reshape(x_fake_, tf.shape(x_real))
            real_output = self.D(x_real, training=True)
            fake_output = self.D(x_fake, training=True)
            cost = self.D_loss(real_output, fake_output)
        grad = t.gradient(cost, self.D.trainable_variables)
        self.D_optimizer.apply_gradients(zip(grad, self.D.trainable_variables))
        return cost    
    
    @tf.function
    def train_step_nv1(self, x_real):
        """Training step for one epoch with 1 generator step and n_disc discriminator step"""
        for i in range(self.n_disc):
            x_real_batch = tf.gather(x_real, i)
            d_loss = self.D_step(x_real_batch)
        g_loss = self.G_step()
        return g_loss, d_loss

    @tf.function
    def train_step_1vn(self, x_real):
        """Training step for one epoch with n_gen generator step and 1 discriminator step"""
        for i in range(self.n_gen):
            g_loss = self.G_step()
        d_loss = self.D_step(x_real)
        return g_loss, d_loss    
    
    def generate_samples(self, batch_size:int, shape:Optional[Tuple[int]]=None):
        """Generates sample using random inputs
        
            Arguments:
                batch_size: int
                    Number of samples to generate.
                shape: (Optional) tuple of int
                    Reshape the output to the given shape.
        """
        z_batch_shape = (batch_size,) + self.latent_dim
        z = tf.random.normal(z_batch_shape)
        samples = self.G(z, training=False)  
        if shape is not None:
            shape = (batch_size,) + shape
            samples = tf.reshape(samples, shape)
        return samples
    
    def predict(self, x):
        """Get predicted output from discriminator
        
            Arguments:
                x: Numpy array, tf.Tensor / any inputs accepted by tensorflow
                    Input data.
            Returns:
                tf.Tensor representing the predicted output from discriminator
        """
        return self.D(x, training=False)

    def train(self, x):
        """Train a GAN model
        
            Arguments:
                x: Numpy array, tf.Tensor / any inputs accepted by tensorflow
                    Input data.        
        """
        self._train_preprocess()
        dataset = self.prepare_dataset(x, batch_size=self.batch_size * self.n_disc,
                                       seed=self.random_state)
        g_metric = tf.keras.metrics.Mean()
        d_metric = tf.keras.metrics.Mean()
        
        self.g_loss_arr = []
        self.d_loss_arr = []
        self.epoch_arr = []
        
        input_shape = x.shape[1:]
        if (self.n_disc == 1) and (self.n_gen == 1):
            train_step = self.train_step_1v1
            input_batch_shape = (self.batch_size,) + input_shape
        elif (self.n_disc > 1) and (self.n_gen == 1):
            train_step = self.train_step_nv1
            input_batch_shape = (self.n_disc, self.batch_size) + input_shape
        elif (self.n_disc == 1) and (self.n_gen > 1):
            train_step = self.train_step_1vn
            input_batch_shape = (self.batch_size,) + input_shape
        else:
            raise RuntimeError("n_disc > 1 and n_gen > 1 are not supported")
            
        for epoch in range(self.epochs):
            for step, x_batch_train_ in enumerate(dataset):
                # for restoring dataset at checkpoint
                if self.start_epoch > epoch:
                    continue
                x_batch_train = tf.reshape(x_batch_train_, input_batch_shape)
                gen_loss, disc_loss = train_step(x_batch_train)
                g_metric(gen_loss)
                d_metric(disc_loss)
                
            self.g_loss_arr.append(g_metric.result().numpy())
            self.d_loss_arr.append(d_metric.result().numpy())
            self.epoch_arr.append(epoch)
            
            self._train_post_epoch(epoch)
            g_metric.reset_states()
            d_metric.reset_states()
        
        self._train_postprocess()
        
    def _train_post_epoch(self, epoch:int, *args, **kwargs):
        super()._train_post_epoch(epoch)
        if self.visualization_interval and ((epoch + 1) % self.visualization_interval == 0):
            self.display_loss_and_image(self.g_loss_arr, self.d_loss_arr, self.epoch_arr)
        
    def enable_visualization(self, image_shape:Tuple[int], n_image:int=16, interval:int=1):
        """Enable visualization of loss curve and generated images
        
            Arguments:
                image_shape: tuple of int
                    Dimensions of the image of the form (rows, cols).
                n_image: int, default=16
                    Number of images to show.
                interval: int, default=1
                    Number of epochs between each update.
        """
        self.image_shape = image_shape
        self.visualization_interval = interval
        self.n_image_to_show = n_image
    
    def disable_visualization(self):
        self.visualization_interval = None
    
    def display_loss_and_image(self, g_loss, d_loss, epochs):
        if clear_output is not None:
            clear_output(wait=True)
        fig = plt.figure(figsize=(16,9))
        size = max(self.n_image_to_show, 1)
        rows = ( size // 4 ) + 1
        gs = gridspec.GridSpec(ncols=8, nrows=rows, figure=fig)
        epoch = epochs[-1]
        # plot loss curve
        ax_loss = plt.subplot(gs[:,:4])
        ax_loss.set_xlim(0, 1.1*epoch)
        ax_loss.plot(epochs, g_loss, label="Generator")
        ax_loss.plot(epochs, d_loss, label="Discriminator")
        ax_loss.set_xlabel('Epoch', fontsize=20)
        ax_loss.set_ylabel('Loss', fontsize=20)
        ax_loss.grid(True)
        ax_loss.legend(fontsize=15)        
        if (self.image_shape is not None):
            images = self.generate_samples(self.n_image_to_show, self.image_shape)
            for i in range(images.shape[0]):
                ax = plt.subplot(gs[i//4, 4 + i%4])
                plt.imshow(images[i])
        if self.checkpoint_dir:
            if not os.path.exists(self.checkpoint_dir):
                os.makedirs(self.checkpoint_dir, exist_ok=True)
            image_path = os.path.join(self.checkpoint_dir, 'image_at_epoch_{:04d}.png'.format(epoch))
            plt.savefig(image_path)
        plt.show()


from typing import List, Union, Optional, Callable, Sequence, Tuple

from quple import ParameterisedCircuit


class IsingCoupling(ParameterisedCircuit):
    '''The efficient SU2 parameterised circuit
    
    Implementation based on the Qiskit library.
    
    The IsingCoupling circuit consists of a layer of single-qubit
    rotation gates and a layer of two-qubit ising coupling gates.
    Such gate set is naturally implemented by trapped ions quantum
    computers.
    
    The IsingCoupling circuit may be used as a variational circuit
    or a model circuit in the PQC layer of a machine learning model.
    It represents a unitary U(θ) parameterised by a set of free 
    parameters θ. The circuit may be repeated several times to 
    increase the number of parameters and hence the complexity 
    of the model. 
    
    It is an example of a strongly entangling circuit which has 
    the advantage of capturing correlations between the data
    features at all ranges for a short range circuit. 
    Reference: https://arxiv.org/pdf/1804.00633.pdf
    
    Args:
      n_qubit: int
          Number of qubits in the circuit
      copies: int
         Number of times the layers are repeated.
      rotation_gates: default=['RY','RZ']
          The rotation gates to be used in the circuit.
      entangle_strategy: Determines how the qubits are connected in an entanglement block.
         If str, it specifies the name of the strategy.
         If callable, it specifies the function to map to an interaction graph.
         If list of str, it specifies the names of a list of strategies. The 
         strategy to use is decided by the current block index. For example, if
         the circuit is building the n-th entanglement block in the entanglement layer,
         then the n-th strategy in the list will be used.
         If list of callable, it specifies the list of functions to map to an interaction
         graph. The function to use is decided by the current block index.
      parameter_symbol: str
         Symbol prefix for circuit parameters.
      name: str
         Name of the circuit.    
    '''    


    def __init__(self, n_qubit: int, copies: int=2, 
    		     rotation_gates: Optional[Union[str, 'cirq.Gate', Callable, 
                                          List[str],List['cirq.Gate'],List[Callable]]]=None,
             ising_gates:Optional[Union[str, 'cirq.Gate', Callable, 
                                      List[str],List['cirq.Gate'],List[Callable]]]=None,
             entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                             List[Callable[[int,int],List[Tuple[int]]]]]]=None,
             parameter_symbol:str='θ', name:str='IsingCoupling', *args, **kwargs):
        rotation_gates = rotation_gates or ['RY', 'RZ']
        ising_gates = ising_gates or ['XX']
        super().__init__(n_qubit=n_qubit, copies=copies,
                         rotation_blocks=rotation_gates,
                         entanglement_blocks=ising_gates,
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         name=name,
                         flatten_circuit=False,
                         final_rotation_layer=True,
                         *args, **kwargs)



import numpy as np
import random

def set_global_random_seed(seed:int=None, tf_seed=True):
    """Set global random seed for various python modules such as random, numpy and tensorflow
    """
    random.seed(seed)
    np.random.seed(seed)
    if tf_seed:
        import tensorflow as tf
        tf.random.set_seed(seed)


from typing import Optional, Union, Tuple, Dict

from quple.models.generative import QGAN

import tensorflow as tf
from tensorflow.keras import Model

class QWGAN(QGAN):
    """Quantum Wasserstein Generative Adversarial Network (QWGAN)
    """    
    def __init__(self, generator:Model,
                 discriminator:Model,
                 latent_dim:Optional[Union[int, Tuple]]=None,
                 n_disc:int=3,
                 epochs:int=100, 
                 batch_size:int=32,
                 optimizer:Optional[Union[str, Dict]]=None,
                 optimizer_kwargs:Optional[Dict]=None,
                 name:str='QGAN',
                 random_state:Optional[int]=None,
                 checkpoint_dir:Optional[str]=None,
                 checkpoint_interval:int=10,
                 checkpoint_max_to_keep:Optional[int]=None):
        """ Creates a QGAN model equipped with a generator and a discriminator neural networks.
            The Wasserstein distance is used as the loss function.
        Args:
            generator: tensorflow.keras.Model
                A keras model representing the generator neural network. It can contain
                both classical and quantum layers.
            discriminator: cirq.Circuit or quple.QuantumCircuit instace
                A keras model representing the discriminator neural network. It can contain
                both classical and quantum layers.
            latent_dim: (Optional) int or tuple of int
                Dimensions of the latent space. It specifies the shape of the input noise
                (drawn from a normal distribution) passed to the generator neural network.
            n_disc: (Optional) int, default=3
                Number of discriminator iterations per generator iteration in each epoch.
            epochs: int, default=100
                Number of epochs for the training.
            batch_size: int, default=10
                Number of training examples used in one iteration. Note that the true batch
                size is batch_size * n_disc so that training samples of size batch_size is
                passed to the discriminator in each discriminator step.
            name: (Optional) string
                Name given to the model.
            random_state: (Optional) int
                The random state used at the beginning of a training
                for reproducible result.
            checkpoint_dir: Optional[string]
                The path to a directory in which to write checkpoints. If None,
                no checkpoint will be saved.
            checkpoint_interval: int, default=10
                Number of epochs between each checkpoint.
            checkpoint_max_to_keep: (Optional) int
        """
        super().__init__(generator=generator,
                 discriminator=discriminator,
                 latent_dim=latent_dim,
                 n_disc=n_disc,
                 epochs=epochs,
                 batch_size=batch_size,
                 optimizer=optimizer,
                 optimizer_kwargs=optimizer_kwargs,
                 name=name,
                 random_state=random_state,
                 checkpoint_dir=checkpoint_dir,
                 checkpoint_interval=checkpoint_interval,
                 checkpoint_max_to_keep=checkpoint_max_to_keep)

    @tf.function
    def D_loss(self, real_output, fake_output):
        """Compute discriminator loss."""
        return tf.reduce_mean(fake_output) - tf.reduce_mean(real_output)
    
    @tf.function
    def G_loss(self, fake_output):
        """Compute generator loss."""
        return -tf.reduce_mean(fake_output)

    @tf.function
    def gradient_penalty(self, f, x_real, x_fake):
        alpha = tf.random.uniform([self.batch_size, 1], 0., 1.)
        diff = x_fake - x_real
        inter = x_real + (alpha * diff)
        with tf.GradientTape() as t:
            t.watch(inter)
            pred = f(inter)
        grad = t.gradient(pred, [inter])[0]
        slopes = tf.sqrt(tf.reduce_sum(tf.square(grad), axis=1))
        gp = tf.reduce_mean((slopes - 1.)**2)
        return gp    


from typing import Sequence
import sympy as sp
import numpy as np

import quple
from quple import TemplateCircuitBlock

class PauliBlock(TemplateCircuitBlock):
    def __init__(self, pauli_string:str, encoding_map=None, global_shift=False):
        self.paulis = pauli_string[::-1]
        self.indices = [i for i, pauli in enumerate(self.paulis) if pauli != 'I']
        self.global_shift = global_shift
        self._num_block_qubits = len(self.paulis)
        self.encoding_map = encoding_map
    
    @staticmethod
    def change_basis(circuit:'quple.ParameterisedCircuit', qubits:Sequence[int],
                     pauli_string:Sequence[str], inverse=False) -> None:
        # do not change basis if only first order pauli operator
        if len(pauli_string) == 1:
            return
        for i, pauli in enumerate(pauli_string):
            if pauli == 'X':
                circuit.H(qubits[i])
            elif pauli == 'Y':
                circuit.RX(-np.pi / 2 if inverse else np.pi / 2, qubits[i])            

    def build(self, circuit:'quple.ParameterisedCircuit', qubits:Sequence[int]):
        # Check if no effective Pauli operations
        if not self.indices:
            return None
        
        PauliBlock.change_basis(circuit, qubits, self.paulis)
        
        qubits_to_entangle = tuple(qubits[i] for i in self.indices)
        # prepare encoded parameters
        from quple.data_encoding import EncodingCircuit
        if isinstance(circuit, EncodingCircuit):
            encoded_value = circuit.encode_parameters(list(qubits), 
                                                      encoding_map=self.encoding_map)            
        elif isinstance(circuit, quple.ParameterisedCircuit):
            encoded_value = circuit.new_param(size=1)
        else:
            raise ValueError('Unknown circuit type: {}'.format(type(circuit)))
        circuit.entangle(qubits_to_entangle)
        params = (encoded_value, qubits[-1])
        
        # do not switch to RZ gate if only first order pauli operator
        if len(self.paulis) == 1:
            if self.global_shift:
                if self.paulis[0] == 'Z':
                    circuit.ZPowGate(*params)
                elif self.paulis[0] == 'X':
                    circuit.XPowGate(*params)
                elif self.paulis[0] == 'Y':
                    circuit.YPowGate(*params)                
            else:
                if self.paulis[0] == 'Z':
                    circuit.RZ(*params)
                elif self.paulis[0] == 'X':
                    circuit.RX(*params)
                elif self.paulis[0] == 'Y':
                    circuit.RY(*params)      
        else:
            if self.global_shift:
                circuit.ZPowGate(*params)
            else:
                circuit.RZ(*params)
        circuit.entangle(qubits_to_entangle, inverse=True)
        
        PauliBlock.change_basis(circuit, qubits, self.paulis, inverse=True)       
    
    @property
    def num_block_qubits(self) -> int:
        return self._num_block_qubits


from typing import Tuple
import operator

import numpy as np
from sklearn.model_selection import train_test_split

def _validate_train_val_test_sizes(train_size, val_size, test_size):
    sizes = [train_size, val_size, test_size]
    dtypes = [np.asarray(size).dtype.kind for size in sizes]
    if all(size==None for size in sizes):
        # set default ratios
        train_size = 0.7
        test_size = 0.15
        train_size = 0.15
    elif all(dtype == 'f' for dtype in dtypes):
        sizes_sum = np.sum(sizes)
        if sizes_sum < 0 or sizes_sum > 1:
            raise ValueError(
            'The sum of train_size, test_size and train_size = {}, '
            'should be within (0,1)'.format(sizes_sum))
    elif not all(dtype == 'i' for dtype in dtypes):
        raise ValueError('train_size={}, val_size={}, test_size={} must '
                         'be all integers, all floats or all None'.format(
                             train_size, val_size, test_size))
        
    return tuple(sizes)
    

def train_val_test_split(x:np.ndarray, y:np.ndarray, train_size=None, val_size=None,
                         test_size=None, stratify=None, shuffle=True, 
                         random_state=None, weight:np.ndarray=None):
    """Split inputs into the training, validation and test set based on sklearn train_test_split
    
    Allowed inputs are lists, numpy arrays, scipy-sparse matrices or pandas dataframes.
    
    Args:
        x, y: sequence of indexables with same length / shape[0]
        train_size: float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the train split. If int, represents the absolute number of train samples. If both train_size, val_size and test_size are None, it will be set to 0.7.
        val_size: float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the validation split. If int, represents the absolute number of validation samples. If both train_size, val_size and test_size are None, it will be set to 0.15.
        test_size: float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split. If int, represents the absolute number of test samples. If both train_size, val_size and test_size are None, it will be set to 0.15. 
        stratify: array-like, default=None
            If not None, data is split in a stratified fashion, using this as the class labels.
        shuffle: bool, default=True
            Whether or not to shuffle the data before splitting. If shuffle=False then stratify must be None.            
        random_state: int or RandomState instance, default=None
            Controls the shuffling applied to the data before applying the split. Pass an int for reproducible output across multiple function calls. 
            
    Returns:
        List containing train-val-test split of inputs.            
    """
    train_size, val_size, test_size = _validate_train_val_test_sizes(train_size, val_size, test_size)
    
    arrays = (x, y) if weight is None else (x, y, weight)

    splitting = train_test_split(*arrays, train_size=train_size,
                                          test_size=val_size+test_size,
                                          shuffle=shuffle,
                                          stratify=stratify,
                                          random_state=random_state)
    if np.asarray(val_size).dtype.kind == 'f' and np.asarray(test_size).dtype.kind == 'f':
        normalizer = 1/(val_size+test_size)
        val_size *=normalizer
        test_size *=normalizer
    if stratify is not None:
        stratify = splitting[3]
    arrays = (splitting[1], splitting[3]) if weight is None else (splitting[1], splitting[3], splitting[5])
    new_splitting = train_test_split(*arrays, train_size=val_size,
                                              test_size=test_size,
                                              shuffle=shuffle,
                                              stratify=stratify,
                                              random_state=random_state)                                                    
    if weight is None:
        return (splitting[0], new_splitting[0], new_splitting[1], splitting[2], new_splitting[2], new_splitting[3])
    else:
        return (splitting[0], new_splitting[0], new_splitting[1], splitting[2], new_splitting[2], new_splitting[3], 
                splitting[4], new_splitting[4], new_splitting[5])

def prepare_train_val_test(x:np.ndarray, y:np.ndarray, train_size=None, val_size=None,
                           test_size=None, preprocessors=None, shuffle=True, stratify=None,
                           random_state=None, weight=None):
    """Prepares training, validation and test datasets from inputs with proprocessing 
    
    Allowed inputs are lists, numpy arrays, scipy-sparse matrices or pandas dataframes.
    
    Args:
        x, y: sequence of indexables with same length / shape[0]
        train_size: float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the train split. If int, represents the absolute number of train samples. If both train_size, val_size and test_size are None, it will be set to 0.7.
        val_size: float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the validation split. If int, represents the absolute number of validation samples. If both train_size, val_size and test_size are None, it will be set to 0.15.
        test_size: float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split. If int, represents the absolute number of test samples. If both train_size, val_size and test_size are None, it will be set to 0.15. 
        stratify: array-like, default=None
            If not None, data is split in a stratified fashion, using this as the class labels.
        shuffle: bool, default=True
            Whether or not to shuffle the data before splitting. If shuffle=False then stratify must be None.            
        random_state: int or RandomState instance, default=None
            Controls the shuffling applied to the data before applying the split. Pass an int for reproducible output across multiple function calls. 
        preprocessors: sequence of sklearn transformers 
            Preprocess data with the given preprocessors.
            
    Returns:
        List containing train-val-test split of preprocessed inputs.
    """
    
    splitting = \
    train_val_test_split(x, y, train_size, val_size, test_size, 
                         shuffle=shuffle, 
                         stratify=stratify,
                         random_state=random_state,
                         weight=weight)
    x_train, x_val, x_test = splitting[0], splitting[1], splitting[2]
    y_train, y_val, y_test = splitting[3], splitting[4], splitting[5]
    if weight is not None:
        weight_train, weight_val, weight_test = splitting[6], splitting[7], splitting[8]
    if preprocessors:
        import sklearn
        for preprocessor in preprocessors:
            if not isinstance(preprocessor, sklearn.base.TransformerMixin):
                raise ValueError('Data preprocessor must be an instance of '
                                 'sklearn.base.TransformerMixin')
            transformer = preprocessor.fit(x_train)
            x_train = transformer.transform(x_train)
            x_val = transformer.transform(x_val)
            x_test = transformer.transform(x_test)
    if weight is None:                                       
        return x_train, x_val, x_test, y_train, y_val, y_test  
    else:
        return x_train, x_val, x_test, y_train, y_val, y_test, weight_train, weight_val, weight_test
    
def rescale_data(*data, val_range:Tuple=(0, 1.)):
    min_value = np.min([np.min(d) for d in data])
    max_value = np.max([np.max(d) for d in data])
    range_min = val_range[0]
    range_max = val_range[1]
    rescaled_data = tuple([(((d-min_value)/(max_value-min_value))*(range_max-range_min))+range_min for d in data])
    if len(rescaled_data) == 1:
        return rescaled_data[0]
    return rescaled_data

def crop_images(images:np.ndarray, dimension:Tuple[int]):
    """Crop the central region of an image (2D array) or multiple images (3D array) to the required dimension
    
    Arguments:
        image: numpy.ndarray
            In case of a 2D array, it represents the pixel values of an image with shape (rows, cols)
            In case of a 3D array, it represents a batch of images with shape (batchsize, rows, cols)
        dimension: Tuple[int]
            A 2-tuple specifying the (width, height) of the cropped frame
    Returns:
        Image(s) cropped to the specified dimension
    """
    if images.ndim == 3:
        dimension = (images.shape[0],) + dimension
    elif images.ndim != 2:
        raise ValueError("image must be a 2D or 3D array")
    start = tuple(map(lambda a, da: a//2-da//2, images.shape, dimension))
    end = tuple(map(operator.add, start, dimension))
    slices = tuple(map(slice, start, end))
    return images[slices]    


from typing import Union, Dict, Any
from itertools import repeat
import numbers
import numpy as np
import sympy
from sympy.core import numbers as sympy_numbers
from sympy.functions.elementary.trigonometric import TrigonometricFunction

import pandas as pd
import tensorflow as tf

from quple.utils.utils import parallel_run

_RecursionFlag = object()

tf_ops_map = {
    sympy.sin: tf.sin,
    sympy.cos: tf.cos,
    sympy.tan: tf.tan,
    sympy.asin: tf.asin,
    sympy.acos: tf.acos,
    sympy.atan: tf.atan,
    sympy.atan2: tf.atan2,
    sympy.cosh: tf.cosh,
    sympy.tanh: tf.tanh,
    sympy.sinh: tf.sinh
}


def resolve_inputs(formulas, symbols, inputs):
    param_dicts = pd.DataFrame(inputs, columns=symbols).to_dict("records")
    return tf.convert_to_tensor(parallel_run(resolve_formulas, repeat(formulas), param_dicts))

def resolve_formulas(formulas, param_dict):
    return [resolve_formula(formula, param_dict) for formula in formulas]

def resolve_formula(formula, param_dict, recursive:bool=True, deep_eval_map=None):
    if not isinstance(param_dict, dict):
        raise ValueError("resolver must be a dictionary mapping the raw symbol"
                         " to the corresponding value")
    # Input is a pass through type, no resolution needed: return early
    value = resolve_value(formula)
    if value is not NotImplemented:
        return value
    
    # Handles 2 cases:
    # formula is a string and maps to a number in the dictionary
    # formula is a symbol and maps to a number in the dictionary
    # in both cases, return it directly.
    if formula in param_dict:
        param_value = param_dict[formula]
        value = resolve_value(param_value)
        if value is not NotImplemented:
            return value
        
    # formula is a string and is not in the dictionary.
    # treat it as a symbol instead.
    if isinstance(formula, str):
        # if the string is in the param_dict as a value, return it.
        # otherwise, try using the symbol instead.
        return resolve_formula(sympy.Symbol(formula), param_dict, recursive)
                               
    # formula is a symbol (sympy.Symbol('a')) and its string maps to a number
    # in the dictionary ({'a': 1.0}).  Return it.
    if isinstance(formula, sympy.Symbol) and formula.name in param_dict:
        param_value = param_dict[formula.name]
        value = resolve_value(param_value)
        if value is not NotImplemented:
            return value
    
    # the following resolves common sympy expressions
    if isinstance(formula, sympy.Add):
        summation = resolve_formula(formula.args[0], param_dict, recursive)
        for addend in formula.args[1:]:
            summation += resolve_formula(addend, param_dict, recursive)
        return summation
    if isinstance(formula, sympy.Mul):
        product = resolve_formula(formula.args[0], param_dict, recursive)
        for factor in formula.args[1:]:
            product *= resolve_formula(factor, param_dict, recursive)
        return product
    # for more complicated operations, need to check whether values are tf.Tensors
    is_tensor = any(isinstance(is_tensor, tf.Tensor) for is_tensor in param_dict.values())
    if isinstance(formula, sympy.Pow) and len(formula.args) == 2:
        if is_tensor:
            return tf.pow(resolve_formula(formula.args[0], param_dict, recursive),
                            resolve_formula(formula.args[1], param_dict, recursive))
        return np.power(resolve_formula(formula.args[0], param_dict, recursive),
                        resolve_formula(formula.args[1], param_dict, recursive))
    # for tf.Tensors sympy subs will not preserve the tf.Tensor, need special treatment
    # for the moment, support for trigonometric function only should suffice
    if is_tensor and isinstance(formula, TrigonometricFunction):
        ops = tf_ops_map.get(type(formula), None)
        if ops is None:
            raise ValueError("unsupported sympy operation: {}".format(type(formula)))
        return ops(resolve_formula(formula.args[0], param_dict, recursive))
                               
    if not isinstance(formula, sympy.Basic):
        # No known way to resolve this variable, return unchanged.
        return formula

    # formula is either a sympy formula or the dictionary maps to a
    # formula.  Use sympy to resolve the value.
    # note that sympy.subs() is slow, so we want to avoid this and
    # only use it for cases that require complicated resolution.
    if not recursive:
        # Resolves one step at a time. For example:
        # a.subs({a: b, b: c}) == b
        value = formula.subs(param_dict, simultaneous=True)
        if value.free_symbols:
            return value
        elif sympy.im(value):
            return complex(value)
        else:
            return float(value)
    
    if deep_eval_map is None:
        deep_eval_map = {}
    # Recursive parameter resolution. We can safely assume that value is a
    # single symbol, since combinations are handled earlier in the method.
    if formula in deep_eval_map:
        value = deep_eval_map[formula]
        if value is not _RecursionFlag:
            return value
        raise RecursionError('Evaluation of {value} indirectly contains itself.')

    # There isn't a full evaluation for 'value' yet. Until it's ready,
    # map value to None to identify loops in component evaluation.
    deep_eval_map[formula] = _RecursionFlag

    value = resolve_formula(formula, param_dict, recursive=False)
    if value == formula:
        deep_eval_map[formula] = value
    else:
        deep_eval_map[formula] = resolve_formula(value, param_dict, recursive)
    return deep_eval_map[formula]                        
    
    
def resolve_value(val: Any):
    if isinstance(val, numbers.Number) and not isinstance(val, sympy.Basic):
        return val
    elif isinstance(val, tf.Tensor):
        return val    
    elif isinstance(val, sympy_numbers.IntegerConstant):
        return val.p
    elif isinstance(val, sympy_numbers.RationalConstant):
        return val.p / val.q
    elif val == sympy.pi:
        return np.pi
    else:
        return NotImplemented


from typing import List, Tuple
from math import sqrt
import itertools


"""
Interaction graphs determine how qubits are connected in terms of multi-qubit operations.
An interaction graph accepts two inputs, n and m. Here n refers to the number of qubits in
the system and m is the number of qubits involved in the multi-qubit interaction
"""
    
def nearest_neighbor(n:int, m:int=2) -> List[Tuple[int]]:
    """The nearest neighbor(linear) interaction
    
    In the nearest neighbor interaction, each qubit is connected to its
    next nearest neighbor in a linear manner. 
    
    Examples:
    # Nearest neighbor interaction graph for 4 qubit system with 2 qubit interaction
    >> nearest_neighbor(n=4, m=2)  
    [(0, 1), (1, 2), (2, 3)]
    # Building a circuit of 4 qubits with a layer of CNOT gates using
    # the nearest neighbor interaction graph
    >> cq = ParameterisedCircuit(n_qubit=4)
    >> cq.add_entanglement_layer(['CNOT'], entangle_strategy='linear')
    >> cq
    (0, 0): ───@───────────
               │
    (0, 1): ───X───@───────
                   │
    (0, 2): ───────X───@───
                       │
    (0, 3): ───────────X───
    
    
    Args:
        n: int
        Number of qubits in the system
        m: int
        Number of qubits involved in the multi-qubit interaction
        
    Returns:
        A list of tuples of qubit indices. Each tuple specifies the indices of the
        qubits that are involved in the multi-qubit interaction. 
    """
    return [tuple(range(i, i+m)) for i in range(n - m + 1)]    

def cyclic(n:int, m:int=2) -> List[Tuple[int]]:
    """The cyclic(circular) interaction
    
    In the cyclic interaction, each qubit is connected to its
    next nearest neighbor in a circular manner.

    Examples:
    # Cyclic interaction graph for 4 qubit system with 2 qubit interaction
    >> cyclic(n=4, m=2) 
    [(0, 1), (1, 2), (2, 3), (3, 0)]
    # Building a circuit of 4 qubits with a layer of CNOT gates using
    # the cyclic interaction graph
    >> cq = ParameterisedCircuit(n_qubit=4)
    >> cq.add_entanglement_layer(['CNOT'], entangle_strategy='cyclic')
    >> cq
    (0, 0): ───@───────────X───
               │           │
    (0, 1): ───X───@───────┼───
                   │       │
    (0, 2): ───────X───@───┼───
                       │   │
    (0, 3): ───────────X───@───
    
    
    Args:
        n: int
        Number of qubits in the system
        m: int
        Number of qubits involved in the multi-qubit interaction
        
    Returns:
        A list of tuples of qubit indices. Each tuple specifies the indices of the
        qubits that are involved in the multi-qubit interaction. 
    """    
    return nearest_neighbor(n,m) + [tuple(range(n - m + 1, n)) + (0,)]

def fully_connected(n:int, m:int=2) -> List[Tuple[int]]:
    """The fully-connected(full) interaction
    
    In the fully-connected interaction, every distinct unordered tuple of m qubits
    are connected exactly once. 
    
    Examples:
    # Fully connected interaction graph for 4 qubit system with 2 qubit interaction
    >> fully_connected(n=4, m=2)  
    [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)]
    # Building a circuit of 4 qubits with a layer of CNOT gates using
    # the fully connected interaction graph
    >> cq = ParameterisedCircuit(n_qubit=4)
    >> cq.add_entanglement_layer(['CNOT'], entangle_strategy='full')
    >> cq
                       ┌──┐
    (0, 0): ───@───@────@─────────────
               │   │    │
    (0, 1): ───X───┼────┼@────@───────
                   │    ││    │
    (0, 2): ───────X────┼X────┼───@───
                        │     │   │
    (0, 3): ────────────X─────X───X───
                       └──┘
    
    Args:
        n: int
        Number of qubits in the system
        m: int
        Number of qubits involved in the multi-qubit interaction
        
    Returns:
        A list of tuples of qubit indices. Each tuple specifies the indices of the
        qubits that are involved in the multi-qubit interaction. 
    """        
    return list(itertools.combinations(list(range(n)), m))

def all_to_all(n:int, m:int=2) -> List[Tuple[int]]:
    """The all-to-all interaction
    
    In the fully-connected interaction, every distinct ordered tuple of m qubits
    are connected.
    
    Examples:
    # All to all interaction graph for 4 qubit system with 2 qubit interaction
    >> all_to_all(n=4, m=2)  
    [(0, 1),
     (0, 2),
     (0, 3),
     (1, 0),
     (1, 2),
     (1, 3),
     (2, 0),
     (2, 1),
     (2, 3),
     (3, 0),
     (3, 1),
     (3, 2)]
    # Building a circuit of 4 qubits with a layer of CNOT gates using
    # the all to all interaction graph
    >> cq = ParameterisedCircuit(n_qubit=4)
    >> cq.add_entanglement_layer(['CNOT'], entangle_strategy='all_to_all')
    >> cq
                                   ┌──┐
    (0, 0): ───@───@───@───X─────────X────────────X───────────
               │   │   │   │         │            │
    (0, 1): ───X───┼───┼───@───@────@┼────X───────┼───X───────
                   │   │       │    ││    │       │   │
    (0, 2): ───────X───┼───────X────┼@────@───@───┼───┼───X───
                       │            │         │   │   │   │
    (0, 3): ───────────X────────────X─────────X───@───@───@───
                                   └──┘

    Args:
        n: int
        Number of qubits in the system
        m: int
        Number of qubits involved in the multi-qubit interaction
        
    Returns:
        A list of tuples of qubit indices. Each tuple specifies the indices of the
        qubits that are involved in the multi-qubit interaction. 
    """        
    return list(itertools.permutations(list(range(n)), m))

def star(n:int, m:int=2) -> List[Tuple[int]]:
    """The star interaction
    
    In the star interaction, the first qubit is connected to every other qubit.
    
    Examples:
    # Star interaction graph for 4 qubit system with 2 qubit interaction
    >> star(n=4, m=2)  
    [(0, 1), (0, 2), (0, 3)]
    # Building a circuit of 4 qubits with a layer of CNOT gates using
    # the fully connected interaction graph
    >> cq = ParameterisedCircuit(n_qubit=4)
    >> cq.add_entanglement_layer(['CNOT'], entangle_strategy='star')
    >> cq
    (0, 0): ───@───@───@───
               │   │   │
    (0, 1): ───X───┼───┼───
                   │   │
    (0, 2): ───────X───┼───
                       │
    (0, 3): ───────────X───
    
    Args:
        n: int
        Number of qubits in the system
        m: int
        Number of qubits involved in the multi-qubit interaction
        
    Returns:
        A list of tuples of qubit indices. Each tuple specifies the indices of the
        qubits that are involved in the multi-qubit interaction. 
    """        
    if m > 2:
        raise ValueError('the connectiviy graphs of the "star" method'
                         ' requires 2 qubits in an interaction unit'
                        ' but {} is given'.format(m))
    if m == 2:
        return [(0, i+1) for i in range(n - m + 1)]
    else:
        return [(i,) for i in range(n)]

def alternate_linear(n:int, m:int=2) -> List[Tuple[int]]:
    """The alternate linear interaction
    
    In the alternate, all neiboring qubits are connected but in an alternating manner
    
    Examples:
    # Alternate linear graph for 5 qubit system with 2 qubit interaction
    >> alternate_linear(n=5, m=2)  
    [(0, 1), (2, 3), (1, 2), (3, 4)]
    # Building a circuit of 5 qubits with a layer of CNOT gates using
    # the alternate linear interaction graph
    >> cq = ParameterisedCircuit(n_qubit=4)
    >> cq.add_entanglement_layer(['CNOT'], entangle_strategy='alternate_linear')
    >> cq
    (0, 0): ───@───────
               │
    (0, 1): ───X───@───
                   │
    (0, 2): ───@───X───
               │
    (0, 3): ───X───@───
                   │
    (0, 4): ───────X───
    
    Args:
        n: int
        Number of qubits in the system
        m: int
        Number of qubits involved in the multi-qubit interaction
        
    Returns:
        A list of tuples of qubit indices. Each tuple specifies the indices of the
        qubits that are involved in the multi-qubit interaction. 
    """       
    if m > 2:
        raise ValueError('The "alternate linear" connectivity graphs requires <= 2 '
                         'qubits in the interaction unit'
                        ' but {} is given'.format(m))
    if n < 2:
        raise ValueError('The "alternate linear" connectivity graphs requires at least '
                         '2 qubits in the circuit')
    if m == 2:
        return [(i, i+1) for i in range(0, n - 1, m)] + [(i+1, i+2) for i in range(0, n - 2, m)]  
    else:
        return [(i,) for i in range(n)]
    
def alternate_cyclic(n:int, m:int=2) -> List[Tuple[int]]:
    """The alternate linear interaction
    
    In the alternate, all neiboring qubits as well as the first and last qubit 
    are connected in an alternating manner
    
    Examples:
    # Alternate cyclic graph for 5 qubit system with 2 qubit interaction
    >> alternate_cyclic(n=5, m=2)  
    [(0, 1), (2, 3), (1, 2), (3, 4), (4, 0)]
    # Building a circuit of 5 qubits with a layer of CNOT gates using
    # the alternate linear interaction graph
    >> cq = ParameterisedCircuit(n_qubit=4)
    >> cq.add_entanglement_layer(['CNOT'], entangle_strategy='alternate_linear')
    >> cq
    (0, 0): ───@───────X───
               │       │
    (0, 1): ───X───@───┼───
                   │   │
    (0, 2): ───@───X───┼───
               │       │
    (0, 3): ───X───@───┼───
                   │   │
    (0, 4): ───────X───@───
    
    Args:
        n: int
        Number of qubits in the system
        m: int
        Number of qubits involved in the multi-qubit interaction
        
    Returns:
        A list of tuples of qubit indices. Each tuple specifies the indices of the
        qubits that are involved in the multi-qubit interaction. 
    """   
    if m > 2:
        raise ValueError('The "alternate linear" connectivity graphs requires <= 2 '
                         'qubits in the interaction unit'
                        ' but {} is given'.format(m))
    if n < 2:
        raise ValueError('The "alternate linear" connectivity graphs requires at least '
                         '2 qubits in the circuit')
    if m == 2:
        return [(i, i+1) for i in range(0, n - 1, m)] + [(i+1, i+2) for i in range(0, n - 2, m)] + [tuple(range(n - m + 1, n)) + (0,)]
    else:
        return [(i,) for i in range(n)]   
    
def filter_mesh(n:int, m:int=2) -> List[Tuple[int]]:   
    #check if number of qubit is a square number
    dimension = sqrt(n)
    if not dimension.is_integer():
        raise ValueError('The "filter_mesh" connectivity graphs requires number of qubits being a square number')
    dimension = int(dimension)
    
    horizontal_filter = [(dimension*i+j, dimension*i+j+1) for i in range(dimension) for j in range(dimension-1)]
    vertical_filter = [(dimension*j+i, dimension*(j+1)+i) for i in range(dimension) for j in range(dimension-1)]
    return horizontal_filter+vertical_filter

def pool_mesh(n:int, m:int=2) -> List[Tuple[int]]:   
    #check if number of qubit is a square number
    dimension = sqrt(n)
    if not dimension.is_integer():
        raise ValueError('The "pool_mesh" connectivity graphs requires number of qubits being a square number')
    dimension = int(dimension)
    if dimension % 2 != 0:
        raise ValueError('The "pool_mesh" connectivity graphs requires number of qubits being a square of an even number')
    
    horizontal_filter = [(dimension*i+j, dimension*i+j+1) for i in range(dimension) for j in range(0, dimension-1, 2)]
    vertical_filter = [(dimension*j+i, dimension*(j+1)+i) for i in range(dimension) for j in range(0, dimension-1, 2)]
    return horizontal_filter+vertical_filter

# dictionary for mapping a str to the corresponding interaction graph
interaction_graph = {
    'nearest_neighbor': nearest_neighbor,
    'linear': nearest_neighbor,
    'cyclic': cyclic,
    'circular': cyclic,
    'full': fully_connected,
    'fully_connected': fully_connected,
    'all_to_all': all_to_all,
    'star': star,
    'alternate_linear': alternate_linear,
    'alternate_cyclic': alternate_cyclic,
    'filter_mesh': filter_mesh,
    'pool_mesh': pool_mesh
}


from typing import List, Union, Optional, Callable, Sequence, Tuple

from quple import ParameterisedCircuit

class RealAmplitudes(ParameterisedCircuit):
    '''The real amplitudes parameterised circuit
    
    Implementation based on the Qiskit library.
    
    The RealAmplitudes circuit consists of a layer
    of Pauli Y rotations acting on each qubit followed by
    a layer of CNOT entanglement on pairs of qubits under
    a given interaction graph. The resultant quantum state
    from the RealAmplitudes circuit will only have real
    amplitudes with 0 complex part. 
    
    The RealAmplitudes circuit may be used as a variational
    circuit or a model circuit in the PQC layer of a machine 
    learning model. It represents a unitary U(θ) parameterised 
    by a set of free  parameters θ. The circuit may be repeated
    several times to  increase the number of parameters and 
    hence the complexity of the model. 
    
    It is an example of a strongly entangling circuit which has 
    the advantage of capturing correlations between the data
    features at all ranges for a short range circuit. 
    Reference: https://arxiv.org/pdf/1804.00633.pdf

    Args:
       n_qubit: int
            Number of qubits in the circuit
       copies: int
           Number of times the layers are repeated.
       entangle_strategy: Determines how the qubits are connected in an entanglement block.
           If str, it specifies the name of the strategy.
           If callable, it specifies the function to map to an interaction graph.
           If list of str, it specifies the names of a list of strategies. The 
           strategy to use is decided by the current block index. For example, if
           the circuit is building the n-th entanglement block in the entanglement layer,
           then the n-th strategy in the list will be used.
           If list of callable, it specifies the list of functions to map to an interaction
           graph. The function to use is decided by the current block index.
       parameter_symbol: str
           Symbol prefix for circuit parameters.
       name: str
           Name of the circuit.
    '''    
    def __init__(self, n_qubit: int, copies: int=2, 
                 entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                 parameter_symbol:str='θ', name:str='RealAmplitudes', *args, **kwargs):
        super().__init__(n_qubit=n_qubit, copies=copies,
                         rotation_blocks='RY',
                         entanglement_blocks='CX',
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         name=name,
                         final_rotation_layer=True,
                         flatten_circuit=False,
                         *args, **kwargs)


import cirq
import numpy as np
import sympy as sp
from itertools import combinations
from sympy import default_sort_key

import quple
from quple.utils.utils import natural_key

#reference: https://github.com/tensorflow/quantum/blob/v0.3.0/tensorflow_quantum/python/util.py
def symbols_in_op(op):
    """Returns the set of symbols associated with a parameterized gate operation.
    
    Arguments:
        op: cirq.Gate
            The parameterised gate operation to find the set of symbols associated with
    
    Returns:
        Set of symbols associated with the parameterized gate operation
    """
    if isinstance(op, cirq.EigenGate):
        return op.exponent.free_symbols

    if isinstance(op, cirq.FSimGate):
        ret = set()
        if isinstance(op.theta, sp.Basic):
            ret |= op.theta.free_symbols
        if isinstance(op.phi, sp.Basic):
            ret |= op.phi.free_symbols
        return ret

    if isinstance(op, cirq.PhasedXPowGate):
        ret = set()
        if isinstance(op.exponent, sp.Basic):
            ret |= op.exponent.free_symbols
        if isinstance(op.phase_exponent, sp.Basic):
            ret |= op.phase_exponent.free_symbols
        return ret

    raise ValueError("Attempted to scan for symbols in circuit with unsupported"
                     " ops inside. Expected op found in tfq.get_supported_gates"
                     " but found: ".format(str(op)))
    
def symbols_in_expr_map(expr_map, to_str=False, sort_key=natural_key):
    """Returns the set of symbols in an expression map
    
    Arguments:
        expr_map: cirq.ExpressionMap
            The expression map to find the set of symbols in
        to_str: boolean, default=False
            Whether to convert symbol to strings
        sort_key: 
            Sort key for the list of symbols
    Returns:
        Set of symbols in the experssion map
    """    
    all_symbols = set()
    for expr in expr_map:
        if isinstance(expr, sp.Basic):
            all_symbols |= expr.free_symbols
    sorted_symbols = sorted(list(all_symbols), key=sort_key)
    if to_str:
        return [str(x) for x in sorted_symbols]
    return sorted_symbols

def get_circuit_unflattened_symbols(circuit:'quple.QuantumCircuit',
                                    to_str=True,
                                    sort_key=natural_key):
    """Returns a list of unflattened parameter symbols in a circuit
    
    Arguments:
        circuit: quple.QuantumCircuit
            The circuit to find the unflattened parameter symbols in
        to_str: boolean, default=True
            Whether to convert symbol to strings
        sort_key:
            Sort key for the list of symbols
    Returns:
        List of unflattened parameter symbols in a circuit
    """      
    if isinstance(circuit, quple.QuantumCircuit):
        expr_map = circuit.expr_map
        if expr_map is not None:
            symbols = quple.symbols_in_expr_map(expr_map, 
                                                to_str=to_str,
                                                sort_key=sort_key)
        else:
            symbols = quple.get_circuit_symbols(circuit,
                                                to_str=to_str,
                                                sort_key=sort_key)
    else:
        symbols = quple.get_circuit_symbols(circuit,
                                            to_str=to_str,
                                            sort_key=sort_key)
    return symbols

#reference: https://github.com/tensorflow/quantum/blob/v0.3.0/tensorflow_quantum/python/util.py
def get_circuit_symbols(circuit, to_str=True, sort_key=natural_key):
    """Returns a list of parameter symbols in a circuit
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to find the associated parameter symbols
        to_str: boolean, default=True
            Whether to convert symbol to strings
        sort_key:
            Sort key for the list of symbols
    Returns:
        A list of symbols in the circuit
    """      
    all_symbols = set()
    for moment in circuit:
        for op in moment:
            if cirq.is_parameterized(op):
                all_symbols |= symbols_in_op(op.gate)
    sorted_symbols = sorted(list(all_symbols), key=sort_key)
    if to_str:
        return [str(x) for x in sorted_symbols]
    return sorted_symbols

def get_circuit_qubits(circuit):
    """Returns a list of qubits in a circuit
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to find the associated qubits
    Returns:
        A list of qubits in the circuit
    """      
    all_qubits = set()
    for moment in circuit:
        for op in moment:
            all_qubits |= set(op._qubits)
    return sorted(list(all_qubits))

def get_circuit_symbols_in_order(circuit, to_str=False):
    """Returns a list of parameter symbols in a circuit in order of creation
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to find the associated parameter symbols
        to_str: boolean, default=False
            Whether to convert the sympy symbols to strings
    Returns:
        A list of symbols in the circuit in order of creation
    """        
    all_symbols = set()
    symbols_in_order = []
    for moment in circuit:
        for op in moment:
            if cirq.is_parameterized(op):
                new_symbols = symbols_in_op(op.gate)
                symbols_in_order += (new_symbols - all_symbols)
                all_symbols |= symbols_in_op(op.gate)
    if to_str:
        symbols_in_order = [str(symbol) for symbol in symbols_in_order]
    return symbols_in_order

def sample_final_states(circuit, samples=1, data=None, backend=None):
    """Samples the final states of a circuit
    
    If the circuit has parameterised gate operation, random values of
    the symbol values in the range (0, 2π) will be assigned to the gate operation.
    If `data` is given, the symbol values will be assigned according to the given data.
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to sample the final states
        samples: int
            Number of samples, default=1
        data: real numbers with shape
            [batch_size, n_params] specifying parameter values to resolve
            into the circuits, following the natural ordering of symbols, e.g. x_1 < x_2 < x_10
    
    Returns:
        A list of sampled final states of the circuit
    """      
    if isinstance(circuit, cirq.Circuit) and not isinstance(circuit, quple.QuantumCircuit):
        circuit = QuantumCircuit.from_cirq(circuit)
    n_symbols = len(circuit.raw_symbols)
    if (n_symbols > 0) and (data is None):
        data = np.random.rand(samples, n_symbols)*2*np.pi
    return circuit.get_state_vectors(data)

def sample_density_matrices(circuit, samples=1, data=None, backend=None):
    """Samples the density matrices of a circuit
    
    If the circuit has parameterised gate operation, random values of
    the symbol values in the range (0, 2π) will be assigned to the gate operation.
    If `data` is given, the symbol values will be assigned according to the given data.
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to sample the density matrices
        samples: int, default=1
            Number of samples
        data: real numbers with shape
            [batch_size, n_params] specifying parameter values to resolve
            into the circuits, following the natural ordering of symbols, e.g. x_1 < x_2 < x_10
    
    Returns:
        A list of sampled density matrices of the circuit
    """       
    final_states = sample_final_states(circuit, samples, data=data, backend=backend)
    density_matrices = [cirq.density_matrix_from_state_vector(fs) for fs in final_states] 
    return density_matrices


def sample_fidelities(circuit, samples=1, data=None, backend=None):
    """Samples the fidelities between two sampled final states of a circuit
    
    Two independent set of samples of circuit final states are first generated.
    The two set of samples are then paired according to the indices of the set.
    The fidelity between the final states in each pair of the samples is then calculated
    to obtain the sampled fidelties of the circuit. 
    
    If the circuit has parameterised gate operation, random values of
    the symbol values in the range (0, 2π) will be assigned to the gate operation.
    If `data` is given, the symbol values will be assigned according to the given data.
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to sample the fidelities
        samples: int, default=1
            Number of samples
        data: real numbers with shape
            [batch_size, n_params] specifying parameter values to resolve
            into the circuits, following the ordering dictated by `symbol_names`
    
    Returns:
        A list of sampled fidelities between two sampled final states of a circuit
    """           
    sample_states_1 = sample_final_states(circuit, samples, data=data, backend=backend)
    sample_states_2 = sample_final_states(circuit, samples, data=data, backend=backend)
    fidelities = []
    for s1, s2 in zip(sample_states_1, sample_states_2):
        fidelities.append(cirq.fidelity(s1, s2))
    return fidelities

def circuit_fidelity_pdf(circuit, samples=3000, bins=100, data=None, backend=None):
    """Returns the binned probability density function from the sampled fidelties of a circuit
    
    If the circuit has parameterised gate operation, random values of
    the symbol values in the range (0, 2π) will be assigned to the gate operation.
    If `data` is given, the symbol values will be assigned according to the given data.    
    
    Arguments: 
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to sample the fidelities
        samples: int, default=3000
            Number of samples
        bins: int, default=100
            Number of bins for the fidelity pdf
        data: real numbers with shape
            [batch_size, n_params] specifying parameter values to resolve
            into the circuits, following the natural ordering of symbols, e.g. x_1 < x_2 < x_10
    Returns:
        A numpy array of the probability density function from the sampled fidelties of a circuit
    """      
    data = np.array(sample_fidelities(circuit, samples, data=data, backend=backend))
    pdf = np.histogram(data, bins=bins, range=(0,1), density=True)[0]
    pdf /= pdf
    return pdf


def get_data_Haar(n_qubit, samples=3000, bins=100):
    """Returns the fidelity values sampled from the Haar distribution
    
    Arguments:
        n_qubit: int
            Number of qubitts
        samples: int, default=3000
            Number of samples
        bins: int, default=100
            Number of bins of the Haar fidelity distribution
    Returns:
        A list of fidelity values sampled from the Haar distribution
    """      
    x = np.linspace(0., 1., bins)
    pdf = get_pdf_Haar(n_qubit, x)
    data = [np.random.choice(x, p=pdf) for _ in range(samples)]
    return data

def get_pdf_Haar(n_qubit, f_values):
    """Returns the Haar probability density function

    Arguments: 
        n_qubit: integer
            Number of qubits
        f_values: list/array of float/int
            A collection of fidelity values to be sampled 
    Returns:
        A numpy array of the Haar probability density function
    """          
    N = 2**n_qubit
    pdf = (N-1)*(1-f_values)**(N-2)
    pdf /= pdf.sum()
    return pdf

def circuit_fidelity_plot(circuit, samples=3000, bins=100, data=None, KL=True, epsilon=1e-10, backend=None):  
    """Returns a plot of the fidelity distribution for the final states sampled from a parameterised circuit overlayed with the Haar fidelity distrubiotn
    
    
    If the circuit has parameterised gate operation, random values of
    the symbol values in the range (0, 2π) will be assigned to the gate operation.
    If `data` is given, the symbol values will be assigned according to the given data.
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to sample the fidelities
        samples: int
            Number of samples
        bins: int, default=100
            Number of bins for the fidelity pdf
        data: real numbers with shape
            [batch_size, n_params] specifying parameter values to resolve
            into the circuits, following the ordering dictated by `symbol_names`
        KL: boolean, default=True
            If True, include the KL divergence between the fidelity distribution from the circuit and the Haar distribution
        epsilon: float, default=1e-10
            Replace a zero bin with a bin of size epsilon to prevent infinity in the KL-divergence
    Returns:
        A plot of the fidelity distribution for the final states sampled from a parameterised circuit overlayed with the Haar fidelity distrubiotn
    """      
    import matplotlib.pyplot as plt
    data_pqc = np.array(sample_fidelities(circuit, samples, data=data, backend=backend))
    #data_Haar = np.linspace(0., 1., samples)
    n_qubit = len(get_circuit_qubits(circuit))
    data_Haar = get_data_Haar(n_qubit, samples, bins)
    plt.clf()
    plt.hist(data_Haar, bins=bins, range=(0.,1.), alpha=0.5, density=True, label='Haar')
    plt.hist(data_pqc, bins=bins, range=(0.,1.), alpha=0.5, density=True, label='PQC')
    plt.legend()
    plt.xlabel('Fidelity')
    plt.ylabel('Probability')
    
    if KL:
        import scipy as sp
        pdf_pqc = np.histogram(data_pqc, bins=bins, range=(0.,1.), density=True)[0]
        pdf_pqc /= pdf_pqc.sum()
        pdf_Haar = get_pdf_Haar(n_qubit, np.linspace(0., 1., bins))
        pdf_Haar = np.array([epsilon if v == 0 else v for v in pdf_Haar])
        Kullback_Leibler_divergence = sp.stats.entropy(pdf_pqc, pdf_Haar)
        plt.title('$D_{KL}=$'+str(Kullback_Leibler_divergence))
    
    return plt

def circuit_expressibility_measure(circuit, samples = 3000, bins=100, data=None, relative=False, epsilon=1e-10, backend=None):
    """Returns the expressibility measure of a parameterised circuit.
    
    The expressibility measure is the KL-divergence between the sampled fidelity distribution of the parameterised circuit and the Haar distribution. 

    Reference: https://arxiv.org/pdf/1905.10876.pdf
    
    If the circuit has parameterised gate operation, random values of
    the symbol values in the range (0, 2π) will be assigned to the gate operation.
    If `data` is given, the symbol values will be assigned according to the given data.
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to find the expressibiliy measure
        samples: int, default=3000
            Number of samples
        bins: int, default=100
            Number of bins for the fidelity pdf
        data: real numbers with shape
            [batch_size, n_params] specifying parameter values to resolve
            into the circuits, following the ordering dictated by `symbol_names`
        relative: boolean, default=False
            If True, calculate the relative KL-divergence
        epsilon: float, default=1e-10
            Replace a zero bin with a bin of size epsilon to prevent infinity in the KL-divergence
    Returns:
        The expressibility measure of a parameterised circuit.
    """       
    import scipy as sp
    data_pqc = np.array(sample_fidelities(circuit, samples, data=data, backend=backend))
    pdf_pqc = np.histogram(data_pqc, bins=bins, range=(0.,1.), density=True)[0]
    pdf_pqc /= pdf_pqc.sum()
    n_qubit = len(get_circuit_qubits(circuit))
    pdf_Haar = get_pdf_Haar(n_qubit, np.linspace(0., 1., bins))
    pdf_Haar = np.array([epsilon if v == 0 else v for v in pdf_Haar])
    Kullback_Leibler_divergence = sp.stats.entropy(pdf_pqc, pdf_Haar)
    expressibility = Kullback_Leibler_divergence
    if relative:
        expressibility_idle_circuit = (2**n_qubit-1)*(np.log(bins))
        expressibility = -np.log(expressibility/expressibility_idle_circuit)
    return expressibility

def Meyer_Wallach_measure(state):
    """Returns the Meyer Wallach measure of a quantum state
    
    Reference: https://arxiv.org/pdf/quant-ph/0305094.pdf
    
    Arguments:
        state: array like
            The quantum state to calculate the Meyer Wallach measure
            
    Returns:
        The Meyer Wallach measure of a quantum state
    """      
    state = np.array(state)
    size = state.shape[0]
    n = int(np.log2(size))
    Q = 0.
    
    def linear_mapping(b:int, j:int) -> np.ndarray:
        keep_indices = [i for i in range(size) if b == ((i & (1 << j))!=0)]
        return state[keep_indices]
    
    def distance(u:np.ndarray, v:np.ndarray) -> float:
        return np.sum(np.abs(np.outer(u,v) - np.outer(v,u))**2)/2
    
    for k in range(n):
        iota_0k = linear_mapping(0, k)
        iota_1k = linear_mapping(1, k)
        Q += distance(iota_0k, iota_1k)
    Q = (4/n)*Q
    return Q
    
        
def circuit_entangling_measure(circuit, samples=200, data=None, backend=None):
    """Returns the entangling measure of a parameterised circuit
    
    The entangling measure of a parameterised circuit is the average Meyer Wallach 
    measure of the sampled final states of the circuit.
    Reference: https://arxiv.org/pdf/1905.10876.pdf
    
    If the circuit has parameterised gate operation, random values of
    the symbol values in the range (0, 2π) will be assigned to the gate operation.
    If `data` is given, the symbol values will be assigned according to the given data.
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to find the entangling measure
        samples: int, default=200
            Number of samples
        data: real numbers with shape
            [batch_size, n_params] specifying parameter values to resolve
            into the circuits, following the ordering dictated by `symbol_names`            
    Returns:
        The entangling measure of a parameterised circuit
    """      
    final_states = sample_final_states(circuit, samples, data=data, backend=backend)
    mw_measures = [Meyer_Wallach_measure(fs) for fs in final_states]
    return np.mean(mw_measures)


def circuit_von_neumann_entropy(circuit, samples=200, data=None, backend=None):
    """Returns the average von Neumann entropy of the sampled density matrices of a parameterised circuit
    
    If the circuit has parameterised gate operation, random values of
    the symbol values in the range (0, 2π) will be assigned to the gate operation.
    If `data` is given, the symbol values will be assigned according to the given data.
    
    Arguments:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to find the von Neumann entropy
        samples: int, default=200
            Number of samples
        data: real numbers with shape
            [batch_size, n_params] specifying parameter values to resolve
            into the circuits, following the ordering dictated by `symbol_names`            
    Returns:
        The average von neumann entropy of the sampled density matrices of a parameterised circuit
    """        
    density_matrices = sample_density_matrices(circuit, samples, data=data, backend=backend)
    von_neumann_entropy = [cirq.von_neumann_entropy(dm) for dm in density_matrices]
    return np.mean(von_neumann_entropy)
    
    
def gradient_variance_test(circuits, op, symbol=None):
    """Performs the gradient variance test for a parameter symbol in a parameterised circuit
    
    Reference: https://www.nature.com/articles/s41467-018-07090-4
    
    Arguments:
        circuits: list of cirq.Circuit, quple.QuantumCircuit
            The circuits to perform the gradient variance test
        op: cirq.Gate
            The gate operation to sample the expectation value from
        symbol: str, default=None
            The parameter symbol which the values are varied in the gradient variance test.
            If None, the first symbol that appears in the circuit will be varied whereas others are fixed
            by some random value.
    Returns:
        Gradient variance of a parameter symbol among the given circuits
    """      
    import tensorflow_quantum as tfq
    import tensorflow as tf
    """Compute the variance of a batch of expectations w.r.t. op on each circuit that 
    contains `symbol`."""
    resolved_circuits = []
    # Resolve irrelevant symbols:
    for circuit in circuits: 
        symbols = get_circuit_symbols(circuit)
        if not len(symbols) > 0:
            raise ValueError('No symbols found in circuit')
        if not symbol:
            symbol = symbols[0]
            symbols = symbols[1:]
        else:
            symbols.remove(symbol)
        resolver = cirq.ParamResolver({s:np.random.uniform() * 2.0 * np.pi for s in symbols})
        resolved_circuits.append(cirq.protocols.resolve_parameters(circuit, resolver))

    # Setup a simple layer to batch compute the expectation gradients.
    expectation = tfq.layers.Expectation()

    # Prep the inputs as tensors
    circuit_tensor = tfq.convert_to_tensor(resolved_circuits)
    n_circuits = len(resolved_circuits)
    values_tensor = tf.convert_to_tensor(
        np.random.uniform(0, 2 * np.pi, (n_circuits, 1)).astype(np.float32))

    # Use TensorFlow GradientTape to track gradients.
    with tf.GradientTape() as g:
        g.watch(values_tensor)
        forward = expectation(circuit_tensor,
                              operators=op,
                              symbol_names=[symbol],
                              symbol_values=values_tensor)

    # Return variance of gradients across all circuits.
    grads = g.gradient(forward, values_tensor)
    grad_var = tf.math.reduce_std(grads, axis=0)
    return grad_var.numpy()[0]


def has_composite_symbols(circuit):
    if not isinstance(circuit, cirq.Circuit):
        raise ValueError("circuit must be a cirq.Circuit object.")
    if (isinstance(circuit, quple.QuantumCircuit)) and (circuit.expr_map is not None):
        circuit = circuit.get_unflattened_circuit()     
    for moment in circuit:
        for op in moment:
            if cirq.is_parameterized(op):
                if len(symbols_in_op(op.gate)) > 1:
                    return True
    return False
'''
    if not isinstance(circuit, quple.QuantumCircuit):
        circuit = quple.QuantumCircuit.from_cirq(circuit)
    # circuit is already flattened
    if circuit.expr_map is not None:
        return set(circuit.symbols) != set(circuit.raw_symbols)
    circuit_symbols = circuit.symbols
    flattened_circuit, _ = cirq.flatten(circuit)
    flattened_circuit_symbols = get_circuit_symbols(flattened_circuit)
    return set(circuit_symbols) != set(flattened_circuit_symbols)
'''


from typing import List, Union, Optional, Callable, Sequence
import numpy as np
import sympy as sp

import cirq

from quple.data_encoding import EncodingCircuit
from quple.circuits.templates.pauli_block import PauliBlock

class GeneralPauliEncoding(EncodingCircuit):
    '''The general Pauli encoding circuit
    
    An encoding circuit consisting of layers of unitary operators of the
    form exp(iψ(x)Σ)H^{⊗n} where ψ is a data encoding function, Σ is a
    generalized Pauli operator from the general Pauli group G_n 
    which is an n-fold tensor product of Pauli operators on n
    qubits, and x = (x_1, . . . , x_n ) are the input features to be encoded.
    
    To encode data of feature dimension n, a set of general Pauli operators 
    are chosen to encode the data into an n qubit circuit. Each Pauli operator
    will contribute to a unitary operation \exp(i\sum_{s\in S}ψ_s(x_s)Σ_s) where
    s is the indices of a subset of all qubit indices S. For a general Pauli 
    operator of order k, s is a tuple of k elements. 
    
    For example, suppose the Pauli operator Σ = Z is used, then S is the set
    {1, 2, ..., n} and the unitary operation is \exp(i\sum_{j=1}^n ψ_j(x_j)Z_j)H^{⊗n}, 
    where Z_j is the Pauli Z operator acting on the j-th qubit. 
    
    If instead the Pauli Operator Σ = Z⊗Z is used, then S is a set of 2-tuple of
    qubit indices determined by the interaction graph. For a fully connected
    graph, S is the set of combinations of 2-qubit pairs. Then the unitary operation
    is \exp(i\sum{s=(j, k) in S}ψ_{j,k}(x_j, x_k)Z_j⊗Z_k)H^{⊗n}
    
    The Pauli encoding uses the circuit-centric design with alternating
    rotation and entanglement layer to achieve a strongly entangling circuit. 
    Ref: https://arxiv.org/pdf/1804.00633.pdf

    Examples:
    >> cq = GeneralPauliEncoding(feature_dimension=3, paulis=['XX'], copies=2)
    >> cq
    (0, 0): ───H───H───@──────────────────────@───H───H───@──────────────────────@───H──────────────────────────────────────
                       │                      │           │                      │
    (0, 1): ───H───H───X───Rz(pi*<x_0*x_1>)───X───H───────┼──────────────────────┼───H───────@──────────────────────@───H───
                                                          │                      │           │                      │
    (0, 2): ───H──────────────────────────────────────H───X───Rz(pi*<x_0*x_2>)───X───H───H───X───Rz(pi*<x_1*x_2>)───X───H───


    Args:
        feature_dimension: int
            Dimension of data to be encoded (=number of qubits in the circuit)
        copies: int
            Number of repetition of the encoding circuit
        encoding_map: function that maps a numpy array to a number
            Data mapping function from R^{feature_dimension} to R
        paulis: list of str
            Pauli operations to be performed on each circuit block
        name: str
            Name of the encoding circuit
    '''    
    def __init__(self, feature_dimension: int,
                 copies:int=2, paulis:List[str] = ['Z', 'ZZ'],
                 encoding_map:Optional[Callable[[np.ndarray], float]] = None,
                 global_shift:bool=False,
                 parameter_symbol:str='x',
                 parameter_scale=sp.pi,
                 name:str='GeneralPauliEncoding',*args, **kwargs):
        '''Creates the general Pauli encoding circuit
        
        Args:
            feature_dimension: int
                Dimension of data to be encoded (=number of qubits in the circuit)
            copies: int
                Number of repetition of the encoding circuit
            paulis: list of str
                Pauli operations to be performed on each circuit block
            encoding_map: function that maps a numpy array to a number
                Data mapping function from R^{feature_dimension} to R
            global_shift: bool
                Include global shift to RZ gate
            name: str
                Name of the encoding circuit
        '''
        GeneralPauliEncoding._validate_paulis(paulis)
        pauli_blocks = [PauliBlock(pauli, global_shift=global_shift) for pauli in paulis]
        super().__init__(feature_dimension, copies=copies, 
                         entanglement_blocks=pauli_blocks,
                         encoding_map=encoding_map,
                         parameter_symbol=parameter_symbol,
                         parameter_scale=parameter_scale,
                         name=name,
                        *args, **kwargs)
        self.paulis = paulis
        
    @staticmethod
    def _validate_paulis(paulis:List[str]):
        for pauli_str in paulis:
            for pauli in pauli_str:
                if pauli not in ['Z','X','Y','I']:
                    raise ValueError('Invalid Pauli operation: {}'.format(pauli))


from functools import reduce
import numpy as np
import sympy as sp

from quple.components.interaction_graphs import cyclic

def trial_0_qiskit(x: np.ndarray) -> float:
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = (x[0]+x[1])/np.pi
    else:
        raise ValueError('The encoding map "trial_0" supports only up to second order feature maps')
    return coeff

class trial_2_qiskit:
    def __init__(self, degree=1):
        self.degree = degree
    def __call__(self, x: np.ndarray):
        if len(x) == 1:
            coeff = x[0]**self.degree
        elif len(x) == 2:
            coeff = (x[0]+x[1])**self.degree/sp.pi
        else:
            raise ValueError('The encoding map "trial_2_qiskit" supports only up to second order feature maps')
        return coeff

class polynomial:
    def __init__(self, degree:int=1, scale_factor=sp.pi, functional=None):
        self.degree = degree
        self.scale_factor = scale_factor
        if functional is None:
            self.functional = lambda x: x
        else:
            self.functional = functional
    def __call__(self, x):
        f_x = [self.functional(x_i) for x_i in x]
        coeff = (reduce(lambda m, n: m + n, f_x)/len(f_x))**self.degree/self.scale_factor
        return coeff
    
class trial_2:
    def __init__(self, degree=1):
        self.degree = degree
    def __call__(self, x: np.ndarray):
        if len(x) == 1:
            coeff = x[0]**self.degree/sp.pi
        elif len(x) == 2:
            coeff = (x[0]+x[1])**self.degree/sp.pi**2
        else:
            raise ValueError('The encoding map "trial_2" supports only up to second order feature maps')
        return coeff

def trial_1(x: np.ndarray) -> float:
    if len(x) == 1:
        coeff = x[0]/sp.pi
    elif len(x) == 2:
        coeff = (x[0]+x[1])/sp.e**2
    elif len(x) == 3:
        coeff = (x[0]+x[1]+x[2])/sp.e**2
    elif len(x) == 4:
        coeff = (x[0]+x[1]+x[2]+x[3])/sp.e**4
    else:
        raise ValueError('The encoding map "trial_0" supports only up to third order feature maps')
    return coeff

def trial_0(x: np.ndarray) -> float:
    if len(x) == 1:
        coeff = x[0]/sp.pi
    elif len(x) == 2:
        coeff = (x[0]+x[1])/sp.pi**2
    elif len(x) == 3:
        coeff = (x[0]+x[1]+x[2])/sp.pi**2
    elif len(x) == 4:
        coeff = (x[0]+x[1]+x[2]+x[3])/sp.pi**4
    else:
        raise ValueError('The encoding map "trial_0" supports only up to third order feature maps')
    return coeff

def qiskit_default(x: np.ndarray) -> float:
    coeff = 2*x[0]/(sp.pi**2)  if len(x) == 1 else 2*reduce(lambda m, n: m * n, (np.pi-x))/(sp.pi**2)
    return coeff

def self_product(x: np.ndarray) -> float:
    """
    Function: \prod_{i=0, n} x_i
    Domain: (-1, +1)
    Range: (-1, +1)

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, x)  
    return coeff

def shifted_self_product(x: np.ndarray) -> float:
    """
    Function: \prod_{i=0, n} (1-x_i)
    Domain: (-1, +1)
    Range: (0, +2)

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = reduce(lambda m, n: m * n, 1-x)/2**(len(x)-1)
    return coeff

def pi_exponent(x: np.ndarray) -> float:
    """
    Function: 
    f(x) = x if n_dim = 1
    f(x) = pi^(|x_0-x_1|^2/8)/pi if n_dim = 2
    Domain: (-1, +1)
    Range: (pi**(-0.5), 1)
    
    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        return x[0]
    elif len(x) == 2:
        return (sp.pi**(abs(x[0]-x[1])**2/8-1))
    else:
        raise ValueError('Encoding function does not allow input of dimension > 2')
        
def cosine_reciprocal(x: np.ndarray) -> float:
    """
    Function: 
    f(x) = x if n_dim = 1
    f(x) = 1/(3*cos(x_0)*cos(x_1)) if n_dim = 2
    Domain: (-1, +1)
    Range: ?
    
    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        return x[0]
    elif len(x) == 2:
        return 1/(3*sp.cos(x[0])*sp.cos(x[1]))
    else:
        raise ValueError('Encoding function does not allow input of dimension > 2')        
    
def cosine_product(x: np.ndarray) -> float:
    """
    Function: f(x) = \prod_{i=0, n} cos(x_i)
    Domain: (-1, +1)
    Range: (-1, +1)?

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    cos_x = [sp.cos(x_i) for x_i in x]
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, cos_x)  
    return coeff

def modified_cosine_product(x: np.ndarray) -> float:
    """
    Function: f(x) = \prod_{i=0, n} cos(pi*(x_i+1)/2)
    Domain: (-1, +1)
    Range: (-1, +1)

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    cos_x = [sp.cos(sp.pi*(x_i+1)/2) for x_i in x]
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, cos_x)  
    return coeff

def modified_sine_cosine_alternate_product(x: np.ndarray) -> float:
    """
    Function: f(x) = sin(pi*x_0)*cos(pi*x_1)*sin(pi*x_2)*...
    Domain: (-1, +1)
    Range: (-1, +1)

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    mix_x = [sp.cos(sp.pi*x[i]) if i%2 else sp.sin(sp.pi*x[i]) for i in range(len(x))]
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, mix_x)  
    return coeff

def sine_cosine_alternate_product(x: np.ndarray) -> float:
    """
    Function: f(x) = sin(x_0)*cos(x_1)*sin(x_2)*...
    Domain: (-1, +1)
    Range: (-1, +1)?

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    mix_x = [sp.cos(x[i]) if i%2 else sp.sin(sp.pi*x[i]) for i in range(len(x))]
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, mix_x)  
    return coeff

def distance_measure(x: np.ndarray) -> float:
    """
    Function: f(x) = (\prod_{i<j} (x_i-x_j))/2**(n_pairs)
    Domain: (-1, +1)
    Range: (-1, +1)

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = (x[0] - x[1])/2
    else:
        pairs = cyclic(len(x))
        y = [x[a]-x[b] for a,b in pairs]
        coeff = reduce(lambda m, n: m * n, y)/2**len(pairs)  
    return coeff

def one_norm_distance(x: np.ndarray) -> float:
    """
    Calculate the one norm distance between the varaibles
    Function: f(x) = (\sum_{i<j} abs(x_i-x_j))/n_pairs
    Domain: (-1, +1)
    Range: (0, +2)

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = np.abs(x[0] - x[1])
    else:
        pairs = cyclic(len(x))
        y = [np.abs(x[a]-x[b]) for a,b in pairs]
        coeff = reduce(lambda m, n: m + n, y)/(len(pairs))
    return coeff

def two_norm_distance(x: np.ndarray) -> float:
    """
    Calculate the two norm distance between the varaibles
    Function: f(x) = (\sum_{i<j} (x_i-x_j)^2/n_pairs)**0.5
    Domain: (-1, +1)
    Range: (0, +2)

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = np.abs(x[0] - x[1])
    else:
        pairs = cyclic(len(x))
        y = [(x[a] - x[b])**2 for a,b in pairs]
        coeff = ((reduce(lambda m, n: m + n, y)/len(pairs))**(1/2))
    return coeff

def cube_sum(x: np.ndarray) -> float:
    """
    Function: f(x) = (\sum_{i=0, n} x_i^3)/n
    Domain: (-1, +1)
    Range: (-1, +1)
    
    Args:
        x: data

    Returns:
        float: the mapped value
    """
    cube_x = [x_i**3 for x_i in x]
    coeff = reduce(lambda m, n: m + n, cube_x)/len(x)
    return coeff   

def arithmetic_mean(x: np.ndarray) -> float:
    """
    Function: f(x) = (\sum_{i=0, n} x_i)/n
    Domain: (-1, +1)
    Range: (-1, +1)
    
    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m + n, x)/len(x) 
    return coeff    
    
def second_moment(x: np.ndarray) -> float:
    """
    Function: f(x) = ((\sum_{i=0, n} (x_i+1)**2)/n)**(1/2)
    Domain: (-1, +1)
    Range: (0, +2)

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    square_x = [(x_i+1)**2 for x_i in x]
    coeff = x[0] if len(x) == 1 else (reduce(lambda m, n: m + n, square_x)/len(x))**0.5
    return coeff   

def exponential_square_sum(x: np.ndarray) -> float:
    """
    Function: f(x) = exp((\sum_{i=0, n} x_i^2/n)-1)*2
    Domain: (-1, +1)
    Range: (2*exp(-1), +2)
    
    Args:
        x: data

    Returns:
        float: the mapped value
    """
    square_x = [x_i**2 for x_i in x]
    coeff = (sp.exp(reduce(lambda m, n: m + n, square_x)/len(x)-1)*2)
    return coeff         

def exponential_cube_sum(x: np.ndarray) -> float:
    """
    Function: f(x) = exp((\sum_{i=0, n} x_i^3/n)-1)*2
    Domain: (-1, +1)
    Range: (2*exp(-2), +2)
    
    Args:
        x: data

    Returns:
        float: the mapped value
    """
    cube_x = [x_i**3 for x_i in x]
    coeff = (sp.exp(reduce(lambda m, n: m + n, cube_x)/len(x)-1)*2)
    return coeff  

encoding_map_registry = {
    'trial_0': trial_0,
    'trial_1': trial_1,
    'poly_1': polynomial(degree=1),
    'poly_2': polynomial(degree=2),
    'poly_3': polynomial(degree=3),
    'poly_4': polynomial(degree=4),
    'qiskit_default': qiskit_default,
    'self_product': self_product,
    'modified_cosine_product': modified_cosine_product, 
    'cosine_product': cosine_product,
    'sine_cosine_alternate_product': sine_cosine_alternate_product,
    'modified_sine_cosine_alternate_product': modified_sine_cosine_alternate_product,
    'distance_measure': distance_measure,
    'one_norm_distance': one_norm_distance,
    'two_norm_distance': two_norm_distance,
    'arithmetic_mean': arithmetic_mean,
    'second_moment': second_moment,
    'cube_sum': cube_sum,
    'exponential_square_sum': exponential_square_sum,
    'exponential_cube_sum': exponential_cube_sum,
    'pi_exponent': pi_exponent,
    'cosine_reciprocal': cosine_reciprocal
}